# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2017 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Simple example that connects to the first Crazyflie found, and then sends the
reboot signals to 6 anchors ID from 0 to 5. The reset signals is sent
10 times in a row to make sure all anchors are reset to bootloader.
"""
import logging
import time
from threading import Thread

import cflib
from cflib.crazyflie import Crazyflie
from cflib.utils import uri_helper
from lpslib.lopoanchor import LoPoAnchor

uri = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')

logging.basicConfig(level=logging.ERROR)


class LpsRebootToBootloader:
    """Example that connects to a Crazyflie and ramps the motors up/down and
    the disconnects"""

    def __init__(self, link_uri):
        """ Initialize and run the example with the specified link_uri """

        self._cf = Crazyflie()

        self._cf.connected.add_callback(self._connected)
        self._cf.disconnected.add_callback(self._disconnected)
        self._cf.connection_failed.add_callback(self._connection_failed)
        self._cf.connection_lost.add_callback(self._connection_lost)

        self._cf.open_link(link_uri)

        print('Connecting to %s' % link_uri)

    def _connected(self, link_uri):
        """ This callback is called form the Crazyflie API when a Crazyflie
        has been connected and the TOCs have been downloaded."""

        # Start a separate thread to do the motor test.
        # Do not hijack the calling thread!
        Thread(target=self._reboot_thread).start()

    def _connection_failed(self, link_uri, msg):
        """Callback when connection initial connection fails (i.e no Crazyflie
        at the specified address)"""
        print('Connection to %s failed: %s' % (link_uri, msg))

    def _connection_lost(self, link_uri, msg):
        """Callback when disconnected after a connection has been made (i.e
        Crazyflie moves out of range)"""
        print('Connection to %s lost: %s' % (link_uri, msg))

    def _disconnected(self, link_uri):
        """Callback when the Crazyflie is disconnected (called in all cases)"""
        print('Disconnected from %s' % link_uri)

    def _reboot_thread(self):

        anchors = LoPoAnchor(self._cf)

        print('Sending reboot signal to all anchors 10 times in a row ...')
        for retry in range(10):
            for anchor_id in range(6):
                anchors.reboot(anchor_id, anchors.REBOOT_TO_BOOTLOADER)
                time.sleep(0.1)

        self._cf.close_link()


if __name__ == '__main__':
    # Initialize the low-level drivers
    cflib.crtp.init_drivers()

    le = LpsRebootToBootloader(uri)



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2017-2018 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
'''
Example of a swarm sharing data and performing a leader-follower scenario
using the motion commander.

The swarm takes off and the drones hover until the follower's local coordinate
system is aligned with the global one. Then, the leader performs its own
trajectory based on commands from the motion commander. The follower is
constantly commanded to keep a defined distance from the leader, meaning that
it is moving towards the leader when their current distance is larger than the
defined one and away from the leader in the opposite scenario.
All movements refer to the local coordinate system of each drone.

This example is intended to work with an absolute positioning system, it has
been tested with the lighthouse positioning system.

This example aims at documenting how to use the collected data to define new
trajectories in real-time. It also indicates how to use the swarm class to
feed the Crazyflies completely different asynchronized trajectories in parallel.
'''
import math
import time

import cflib.crtp
from cflib.crazyflie.log import LogConfig
from cflib.crazyflie.swarm import CachedCfFactory
from cflib.crazyflie.swarm import Swarm
from cflib.positioning.motion_commander import MotionCommander

# Change uris according to your setup
# URIs in a swarm using the same radio must also be on the same channel
URI1 = 'radio://0/80/2M/E7E7E7E7E7'  # Follower
URI2 = 'radio://0/80/2M/E7E7E7E7E8'  # Leader


DEFAULT_HEIGHT = 0.75
DEFAULT_VELOCITY = 0.5
x1 = [0]
y1 = [0]
z1 = [0]
x2 = [0]
y2 = [0]
z2 = [0]
yaw1 = [0]
yaw2 = [0]
d = 0

# List of URIs
uris = {
    URI1,
    URI2,
}


def wait_for_param_download(scf):
    while not scf.cf.param.is_updated:
        time.sleep(1.0)
    print('Parameters downloaded for', scf.cf.link_uri)


def arm(scf):
    scf.cf.platform.send_arming_request(True)
    time.sleep(1.0)


def pos_to_vel(x1, y1, x2, y2, dist):
    '''
    This function takes two points on the x-y plane and outputs
    two components of the velocity vector: one along the x-axis
    and one along the y-axis. The combined vector represents the
    total velocity, pointing from point 1 to point 2, with a
    magnitude equal to the DEFAULT_VELOCITY. These 2 velocity
    vectors are meant to be used by the motion commander.
    The distance between them is taken as an argument because it
    is constanlty calculated by position_callback().
    '''
    if dist == 0:
        Vx = 0
        Vy = 0
    else:
        Vx = DEFAULT_VELOCITY * (x2-x1)/dist
        Vy = DEFAULT_VELOCITY * (y2-y1)/dist
    return Vx, Vy


def position_callback(uri, data):
    global d
    if uri == URI1:  # Follower
        x1.append(data['stateEstimate.x'])
        y1.append(data['stateEstimate.y'])
        z1.append(data['stateEstimate.z'])
        yaw1.append(data['stateEstimate.yaw'])
    elif uri == URI2:  # Leader
        x2.append(data['stateEstimate.x'])
        y2.append(data['stateEstimate.y'])
        z2.append(data['stateEstimate.z'])
        yaw2.append(data['stateEstimate.yaw'])

    d = math.sqrt(pow((x1[-1]-x2[-1]), 2)+pow((y1[-1]-y2[-1]), 2))


def start_position_printing(scf):
    log_conf1 = LogConfig(name='Position', period_in_ms=10)
    log_conf1.add_variable('stateEstimate.x', 'float')
    log_conf1.add_variable('stateEstimate.y', 'float')
    log_conf1.add_variable('stateEstimate.z', 'float')
    log_conf1.add_variable('stateEstimate.yaw', 'float')
    scf.cf.log.add_config(log_conf1)
    log_conf1.data_received_cb.add_callback(lambda _timestamp, data, _logconf: position_callback(scf.cf.link_uri, data))
    log_conf1.start()


def leader_follower(scf):
    r_min = 0.8  # The minimum distance between the 2 drones
    r_max = 1.0  # The maximum distance between the 2 drones
    with MotionCommander(scf, default_height=DEFAULT_HEIGHT) as mc:

        # The follower turns until it is aligned with the global coordinate system
        while abs(yaw1[-1]) > 2:
            if scf.__dict__['_link_uri'] == URI1:  # Follower
                if yaw1[-1] > 0:
                    mc.start_turn_right(36 if abs(yaw1[-1]) > 15 else 9)
                elif yaw1[-1] < 0:
                    mc.start_turn_left(36 if abs(yaw1[-1]) > 15 else 9)

            elif scf.__dict__['_link_uri'] == URI2:  # Leader
                mc.stop()
            time.sleep(0.005)

        time.sleep(0.5)

        start_time = time.time()
        # Define the flight time after the follower is aligned
        end_time = time.time() + 20

        while time.time() < end_time:

            if scf.__dict__['_link_uri'] == URI1:  # Follower
                if d > r_max:
                    cmd_vel_x, cmd_vel_y = pos_to_vel(x1[-1], y1[-1], x2[-1], y2[-1], d)
                elif d >= r_min and d <= r_max:
                    cmd_vel_x = 0
                    cmd_vel_y = 0
                elif d < r_min:
                    opp_cmd_vel_x, opp_cmd_vel_y = pos_to_vel(x1[-1], y1[-1], x2[-1], y2[-1], d)
                    cmd_vel_x = -opp_cmd_vel_x
                    cmd_vel_y = -opp_cmd_vel_y

                mc.start_linear_motion(cmd_vel_x, cmd_vel_y, 0)

            elif scf.__dict__['_link_uri'] == URI2:  # Leader
                # Define the sequence of the leader
                if time.time() - start_time < 3:
                    mc.start_forward(DEFAULT_VELOCITY)
                elif time.time() - start_time < 6:
                    mc.start_back(DEFAULT_VELOCITY)
                elif time.time() - start_time < 20:
                    mc.start_circle_right(0.9, DEFAULT_VELOCITY)
                else:
                    mc.stop()

            time.sleep(0.005)
        mc.land()


if __name__ == '__main__':
    cflib.crtp.init_drivers()

    factory = CachedCfFactory(rw_cache='./cache')
    with Swarm(uris, factory=factory) as swarm:

        swarm.reset_estimators()

        swarm.parallel_safe(arm)

        print('Waiting for parameters to be downloaded...')
        swarm.parallel_safe(wait_for_param_download)

        time.sleep(1)

        swarm.parallel_safe(start_position_printing)
        time.sleep(0.5)

        swarm.parallel_safe(leader_follower)
        time.sleep(1)



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2021 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Example to show the use of persistent parameters.

* All persistent parameters are fectched and the current state is printed out.
* The LED ring effect is set to a new value and stored.
* The LED ring effect persisted value is cleared.

Note: this script will change the value of the LED ring.effect parameter
"""
import logging
import sys
from threading import Event

import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.utils import uri_helper

# uri = uri_helper.uri_from_env(default='usb://0')
uri = uri_helper.uri_from_env(default='radio://0/30/2M/E7E7E7E7E7')

# Only output errors from the logging framework
logging.basicConfig(level=logging.ERROR)


def get_persistent_state(cf, complete_param_name):
    wait_for_callback_event = Event()

    def state_callback(complete_name, state):
        print(f'{complete_name}: {state}')
        wait_for_callback_event.set()

    cf.param.persistent_get_state(complete_param_name, state_callback)
    wait_for_callback_event.wait()


def persist_parameter(cf, complete_param_name):
    wait_for_callback_event = Event()

    def is_stored_callback(complete_name, success):
        if success:
            print(f'Persisted {complete_name}!')
        else:
            print(f'Failed to persist {complete_name}!')
        wait_for_callback_event.set()

    cf.param.persistent_store(complete_param_name, callback=is_stored_callback)
    wait_for_callback_event.wait()


def clear_persistent_parameter(cf, complete_param_name):
    wait_for_callback_event = Event()

    def is_stored_cleared(complete_name, success):
        if success:
            print(f'Cleared {complete_name}!')
        else:
            print(f'Failed to clear {complete_name}!')
        wait_for_callback_event.set()

    cf.param.persistent_clear(complete_param_name, callback=is_stored_cleared)
    wait_for_callback_event.wait()


def get_all_persistent_param_names(cf):
    persistent_params = []
    for group_name, params in cf.param.toc.toc.items():
        for param_name, element in params.items():
            if element.is_persistent():
                complete_name = group_name + '.' + param_name
                persistent_params.append(complete_name)

    return persistent_params


if __name__ == '__main__':
    # Initialize the low-level drivers
    cflib.crtp.init_drivers()

    cf = Crazyflie(rw_cache='./cache')

    with SyncCrazyflie(uri, cf=cf) as scf:
        # Get the names of all parameters that can be persisted
        persistent_params = get_all_persistent_param_names(scf.cf)

        if not persistent_params:
            print('No persistent parameters')
            sys.exit(0)

        # Get the state of all persistent parameters
        print('-- All persistent parameters --')
        for complete_name in persistent_params:
            get_persistent_state(scf.cf, complete_name)

        print()
        param_name = 'ring.effect'
        param_value = 10

        print(f'Set parameter {param_name} to {param_value}')
        scf.cf.param.set_value(param_name, param_value)

        print()
        print('Store the new value in persistent memory')
        persist_parameter(scf.cf, param_name)

        print('The new state is:')
        get_persistent_state(scf.cf, param_name)

        print()
        print('Clear the persisted parameter')
        clear_persistent_parameter(scf.cf, param_name)

        print('The new state is:')
        get_persistent_state(scf.cf, param_name)



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2017 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
import logging
import sys
import time
from threading import Event

import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.log import LogConfig
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.positioning.motion_commander import MotionCommander
from cflib.utils import uri_helper

URI = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')

DEFAULT_HEIGHT = 0.5
BOX_LIMIT = 0.5

deck_attached_event = Event()

logging.basicConfig(level=logging.ERROR)

position_estimate = [0, 0]


def move_box_limit(scf):
    with MotionCommander(scf, default_height=DEFAULT_HEIGHT) as mc:
        body_x_cmd = 0.2
        body_y_cmd = 0.1
        max_vel = 0.2

        while (1):
            '''if position_estimate[0] > BOX_LIMIT:
                mc.start_back()
            elif position_estimate[0] < -BOX_LIMIT:
                mc.start_forward()
            '''

            if position_estimate[0] > BOX_LIMIT:
                body_x_cmd = -max_vel
            elif position_estimate[0] < -BOX_LIMIT:
                body_x_cmd = max_vel
            if position_estimate[1] > BOX_LIMIT:
                body_y_cmd = -max_vel
            elif position_estimate[1] < -BOX_LIMIT:
                body_y_cmd = max_vel

            mc.start_linear_motion(body_x_cmd, body_y_cmd, 0)

            time.sleep(0.1)


def move_linear_simple(scf):
    with MotionCommander(scf, default_height=DEFAULT_HEIGHT) as mc:
        time.sleep(1)
        mc.forward(0.5)
        time.sleep(1)
        mc.turn_left(180)
        time.sleep(1)
        mc.forward(0.5)
        time.sleep(1)


def take_off_simple(scf):
    with MotionCommander(scf, default_height=DEFAULT_HEIGHT) as mc:
        time.sleep(3)
        mc.stop()


def log_pos_callback(timestamp, data, logconf):
    print(data)
    position_estimate[0] = data['stateEstimate.x']
    position_estimate[1] = data['stateEstimate.y']


def param_deck_flow(_, value_str):
    value = int(value_str)
    print(value)
    if value:
        deck_attached_event.set()
        print('Deck is attached!')
    else:
        print('Deck is NOT attached!')


if __name__ == '__main__':
    cflib.crtp.init_drivers()

    with SyncCrazyflie(URI, cf=Crazyflie(rw_cache='./cache')) as scf:

        scf.cf.param.add_update_callback(group='deck', name='bcFlow2',
                                         cb=param_deck_flow)
        time.sleep(1)

        logconf = LogConfig(name='Position', period_in_ms=10)
        logconf.add_variable('stateEstimate.x', 'float')
        logconf.add_variable('stateEstimate.y', 'float')
        scf.cf.log.add_config(logconf)
        logconf.data_received_cb.add_callback(log_pos_callback)

        if not deck_attached_event.wait(timeout=5):
            print('No flow deck detected!')
            sys.exit(1)

        # Arm the Crazyflie
        scf.cf.platform.send_arming_request(True)
        time.sleep(1.0)

        logconf.start()

        take_off_simple(scf)
        # move_linear_simple(scf)
        # move_box_limit(scf)
        logconf.stop()



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2023 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Shows how to send full state control setpoints to the Crazyflie
"""
import time

from scipy.spatial.transform import Rotation

import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.log import LogConfig
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.utils import uri_helper
from cflib.utils.reset_estimator import reset_estimator

# URI to the Crazyflie to connect to
uri = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')


def quaternion_from_euler(roll: float, pitch: float, yaw: float):
    """Convert Euler angles to quaternion

    Args:
        roll (float): roll, in radians
        pitch (float): pitch, in radians
        yaw (float): yaw, in radians

    Returns:
        array: the quaternion [x, y, z, w]
    """
    return Rotation.from_euler(seq='xyz', angles=(roll, pitch, yaw), degrees=False).as_quat()


def position_callback(timestamp, data, logconf):
    x = data['kalman.stateX']
    y = data['kalman.stateY']
    z = data['kalman.stateZ']
    print('pos: ({}, {}, {})'.format(x, y, z))


def start_position_printing(scf):
    log_conf = LogConfig(name='Position', period_in_ms=500)
    log_conf.add_variable('kalman.stateX', 'float')
    log_conf.add_variable('kalman.stateY', 'float')
    log_conf.add_variable('kalman.stateZ', 'float')

    scf.cf.log.add_config(log_conf)
    log_conf.data_received_cb.add_callback(position_callback)
    log_conf.start()


def send_setpoint(cf, duration, pos, vel, acc, orientation, rollrate, pitchrate, yawrate):
    # Set points must be sent continuously to the Crazyflie, if not it will think that connection is lost
    end_time = time.time() + duration
    while time.time() < end_time:
        cf.commander.send_full_state_setpoint(pos, vel, acc, orientation, rollrate, pitchrate, yawrate)
        time.sleep(0.2)


def run_sequence(scf):
    cf = scf.cf

    # Set to mellinger controller
    # cf.param.set_value('stabilizer.controller', '2')

    # Arm the Crazyflie
    cf.platform.send_arming_request(True)
    time.sleep(1.0)

    print('takeoff')
    send_setpoint(cf, 4.0,
                  [0.0, 0.0, 1.0],
                  [0.0, 0.0, 0.0],
                  [0.0, 0.0, 0.0],
                  quaternion_from_euler(0.0, 0.0, 0.0),
                  0.0, 0.0, 0.0)

    send_setpoint(cf, 2.0,
                  [0.0, 0.0, 1.0],
                  [0.0, 0.0, 0.0],
                  [0.0, 0.0, 0.0],
                  quaternion_from_euler(0.0, 0.0, 0.7),
                  0.0, 0.0, 0.0)
    print('land')
    send_setpoint(cf, 2.0,
                  [0.0, 0.0, 0.1],
                  [0.0, 0.0, 0.0],
                  [0.0, 0.0, 0.0],
                  quaternion_from_euler(0.0, 0.0, 0.0),
                  0.0, 0.0, 0.0)

    cf.commander.send_stop_setpoint()
    # Hand control over to the high level commander to avoid timeout and locking of the Crazyflie
    cf.commander.send_notify_setpoint_stop()

    # Make sure that the last packet leaves before the link is closed
    # since the message queue is not flushed before closing
    time.sleep(0.1)


def _stab_log_data(timestamp, data, logconf):
    print('roll: {}, pitch: {}, yaw: {}'.format(data['controller.roll'],
                                                data['controller.pitch'],
                                                data['controller.yaw']))
    print('ctrltarget.x: {}, ctrltarget.y: {}, ctrltarget.z: {}'.format(data['ctrltarget.x'],
                                                                        data['ctrltarget.y'],
                                                                        data['ctrltarget.z']))


def set_up_logging(scf):
    _lg_stab = LogConfig(name='Stabilizer', period_in_ms=500)
    _lg_stab.add_variable('controller.roll', 'float')
    _lg_stab.add_variable('controller.pitch', 'float')
    _lg_stab.add_variable('controller.yaw', 'float')
    _lg_stab.add_variable('ctrltarget.x', 'float')
    _lg_stab.add_variable('ctrltarget.y', 'float')
    _lg_stab.add_variable('ctrltarget.z', 'float')

    scf.cf.log.add_config(_lg_stab)
    _lg_stab.data_received_cb.add_callback(_stab_log_data)
    _lg_stab.start()


if __name__ == '__main__':
    cflib.crtp.init_drivers()

    with SyncCrazyflie(uri, cf=Crazyflie(rw_cache='./cache')) as scf:
        # set_up_logging(scf)
        reset_estimator(scf)
        run_sequence(scf)



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
import unittest

from cflib.utils.callbacks import Caller


class CallerTest(unittest.TestCase):

    def setUp(self):
        self.callback_count = 0
        self.sut = Caller()

    def test_that_callback_is_added(self):
        # Fixture

        # Test
        self.sut.add_callback(self._callback)

        # Assert
        self.sut.call()
        self.assertEqual(1, self.callback_count)

    def test_that_callback_is_added_only_one_time(self):
        # Fixture

        # Test
        self.sut.add_callback(self._callback)
        self.sut.add_callback(self._callback)

        # Assert
        self.sut.call()
        self.assertEqual(1, self.callback_count)

    def test_that_multiple_callbacks_are_added(self):
        # Fixture

        # Test
        self.sut.add_callback(self._callback)
        self.sut.add_callback(self._callback2)

        # Assert
        self.sut.call()
        self.assertEqual(2, self.callback_count)

    def test_that_callback_is_removed(self):
        # Fixture
        self.sut.add_callback(self._callback)

        # Test
        self.sut.remove_callback(self._callback)

        # Assert
        self.sut.call()
        self.assertEqual(0, self.callback_count)

    def test_that_callback_is_called_with_arguments(self):
        # Fixture
        self.sut.add_callback(self._callback_with_args)

        # Test
        self.sut.call('The token')

        # Assert
        self.assertEqual('The token', self.callback_token)

    def _callback(self):
        self.callback_count += 1

    def _callback2(self):
        self.callback_count += 1

    def _callback_with_args(self, token):
        self.callback_token = token



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2024 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Example to show how to write several persistent parameters from a yaml file.
The params in the file should be formatted like this;

params:
  activeMarker.back:
    default_value: 3
    is_stored: true
    stored_value: 30
type: persistent_param_state
version: '1'
"""
import argparse
import logging

import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.utils import uri_helper
from cflib.utils.param_file_helper import ParamFileHelper


uri = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')

# Only output errors from the logging framework
logging.basicConfig(level=logging.ERROR)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--file', type=str, help='The yaml file containing the arguments. ')
    args = parser.parse_args()

    cflib.crtp.init_drivers()

    cf = Crazyflie(rw_cache='./cache')
    with SyncCrazyflie(uri, cf=cf) as scf:
        writer = ParamFileHelper(scf.cf)
        writer.store_params_from_file(args.file)



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2023 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
import unittest

import numpy as np

from cflib.utils.encoding import compress_quaternion
from cflib.utils.encoding import decompress_quaternion


class EncodingTest(unittest.TestCase):

    def test_compress_decompress(self):
        # Fixture
        expected = self._normalize_quat([1, 2, 3, 4])

        # Test
        compressed = compress_quaternion(expected)
        actual = decompress_quaternion(compressed)

        # Assert
        np.testing.assert_allclose(actual, expected, 0.001)

    def test_compress_decompress_not_normalized(self):
        # Fixture
        quat = [1, 2, 3, 4]
        expected = self._normalize_quat(quat)

        # Test
        compressed = compress_quaternion(quat)
        actual = decompress_quaternion(compressed)

        # Assert
        np.testing.assert_allclose(actual, expected, 0.001)

    def test_other_largest_component(self):
        # Fixture
        quat = [5, 10, 3, 4]
        expected = self._normalize_quat(quat)

        # Test
        compressed = compress_quaternion(quat)
        actual = decompress_quaternion(compressed)

        # Assert
        np.testing.assert_allclose(actual, expected, 0.001)

    def _normalize_quat(self, quat):
        quat = np.array(quat)
        return quat / np.linalg.norm(quat)



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2019 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Simple example of a swarm using the High level commander.

The swarm takes off and flies a synchronous square shape before landing.
The trajectories are relative to the starting positions and the Crazyflies can
be at any position on the floor when the script is started.

This example is intended to work with any absolute positioning system.
It aims at documenting how to use the High Level Commander together with
the Swarm class.
"""
import time

import cflib.crtp
from cflib.crazyflie.swarm import CachedCfFactory
from cflib.crazyflie.swarm import Swarm


def activate_mellinger_controller(scf, use_mellinger):
    controller = 1
    if use_mellinger:
        controller = 2
    scf.cf.param.set_value('stabilizer.controller', controller)


def arm(scf):
    scf.cf.platform.send_arming_request(True)
    time.sleep(1.0)


def run_shared_sequence(scf):
    activate_mellinger_controller(scf, False)

    box_size = 1
    flight_time = 2

    commander = scf.cf.high_level_commander

    commander.takeoff(1.0, 2.0)
    time.sleep(3)

    commander.go_to(box_size, 0, 0, 0, flight_time, relative=True)
    time.sleep(flight_time)

    commander.go_to(0, box_size, 0, 0, flight_time, relative=True)
    time.sleep(flight_time)

    commander.go_to(-box_size, 0, 0, 0, flight_time, relative=True)
    time.sleep(flight_time)

    commander.go_to(0, -box_size, 0, 0, flight_time, relative=True)
    time.sleep(flight_time)

    commander.land(0.0, 2.0)
    time.sleep(2)

    commander.stop()


uris = {
    'radio://0/30/2M/E7E7E7E711',
    'radio://0/30/2M/E7E7E7E712',
    # Add more URIs if you want more copters in the swarm
    # URIs in a swarm using the same radio must also be on the same channel
}

if __name__ == '__main__':
    cflib.crtp.init_drivers()
    factory = CachedCfFactory(rw_cache='./cache')
    with Swarm(uris, factory=factory) as swarm:
        swarm.reset_estimators()
        swarm.parallel_safe(arm)
        swarm.parallel_safe(run_shared_sequence)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2023 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
import struct

import numpy as np


# Code from davidejones at https://gamedev.stackexchange.com/a/28756
def fp16_to_float(float16):
    s = int((float16 >> 15) & 0x00000001)    # sign
    e = int((float16 >> 10) & 0x0000001f)    # exponent
    f = int(float16 & 0x000003ff)            # fraction

    if e == 0:
        if f == 0:
            return int(s << 31)
        else:
            while not (f & 0x00000400):
                f <<= 1
                e -= 1
            e += 1
            f &= ~0x00000400
            # print(s,e,f)
    elif e == 31:
        if f == 0:
            return int((s << 31) | 0x7f800000)
        else:
            return int((s << 31) | 0x7f800000 | (f << 13))

    e += 127 - 15
    f <<= 13
    result = int((s << 31) | (e << 23) | f)
    return struct.unpack('f', struct.pack('I', result))[0]


def decompress_quaternion(comp):
    """Decompress a quaternion

    see quatcompress.h in the firmware for definitions

    Args:
        comp int: A 32-bit number

    Returns:
        np array: q = [x, y, z, w]
    """
    q = np.zeros(4)
    mask = (1 << 9) - 1
    i_largest = comp >> 30
    sum_squares = 0
    for i in range(3, -1, -1):
        if i != i_largest:
            mag = comp & mask
            negbit = (comp >> 9) & 0x1
            comp = comp >> 10
            q[i] = mag / mask / np.sqrt(2)
            if negbit == 1:
                q[i] = -q[i]
            sum_squares += q[i] * q[i]
    q[i_largest] = np.sqrt(1.0 - sum_squares)
    return q


def compress_quaternion(quat):
    """Compress a quaternion.
    assumes input quaternion is normalized. will fail if not.

    see quatcompress.h in firmware the for definitions

    Args:
        quat : An array of floats representing a quaternion [x, y, z, w]

    Returns: 32-bit integer
    """
    # Normalize the quaternion
    quat_n = np.array(quat) / np.linalg.norm(quat)

    i_largest = 0
    for i in range(1, 4):
        if abs(quat_n[i]) > abs(quat_n[i_largest]):
            i_largest = i
    negate = quat_n[i_largest] < 0

    M_SQRT1_2 = 1.0 / np.sqrt(2)

    comp = i_largest
    for i in range(4):
        if i != i_largest:
            negbit = int((quat_n[i] < 0) ^ negate)
            mag = int(((1 << 9) - 1) * (abs(quat_n[i]) / M_SQRT1_2) + 0.5)
            comp = (comp << 10) | (negbit << 9) | mag

    return comp



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2021 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Crazyflie driver using the cflinkcpp implementation.

This driver is used to communicate over the radio or USB.
"""
import logging
import threading

import cflinkcpp

from .crtpstack import CRTPPacket
from cflib.crtp.crtpdriver import CRTPDriver
from cflib.crtp.radio_link_statistics import RadioLinkStatistics

__author__ = 'Bitcraze AB'
__all__ = ['CfLinkCppDriver']

logger = logging.getLogger(__name__)


class CfLinkCppDriver(CRTPDriver):
    """ cflinkcpp driver """

    def __init__(self):
        """Driver constructor. Throw an exception if the driver is unable to
        open the URI
        """
        self.uri = ''

        # cflinkcpp resends packets internally
        self.needs_resending = False

        self._connection = None
        self._radio_link_statistics = RadioLinkStatistics()

    def connect(self, uri, radio_link_statistics_callback, link_error_callback):
        """Connect the driver to a specified URI

        @param uri Uri of the link to open
        @param radio_link_statistics_callback Callback to report radio link statistics
        @param link_error_callback Callback to report errors (will result in
               disconnection)
        """

        self._connection = cflinkcpp.Connection(uri)
        self.uri = uri
        self._radio_link_statistics_callback = radio_link_statistics_callback
        self._link_error_callback = link_error_callback

        if uri.startswith('radio://') and radio_link_statistics_callback is not None:
            self._last_connection_stats = self._connection.statistics
            self._recompute_link_quality_timer()

    def send_packet(self, pk):
        """Send a CRTP packet"""
        nativePk = cflinkcpp.Packet()
        nativePk.port = pk.port
        nativePk.channel = pk.channel
        nativePk.payload = bytes(pk.data)
        try:
            self._connection.send(nativePk)
        except Exception as e:
            if self._link_error_callback is not None:
                import traceback
                self._link_error_callback(
                    'Error communicating! Perhaps your device has been unplugged?\n'
                    'Exception:{}\n\n{}'.format(e, traceback.format_exc()))

    def receive_packet(self, wait=0):
        """Receive a CRTP packet.

        @param wait The time to wait for a packet in second. -1 means forever

        @return One CRTP packet or None if no packet has been received.
        """
        if wait < 0:
            # Since we block in the native lib, break up infinity timeouts into smaller
            # pieces, so Ctrl+C keeps working as expected
            timeout = 100
        elif wait == 0:
            timeout = 1
        else:
            timeout = int(wait*1000)

        try:
            while True:
                nativePk = self._connection.recv(timeout=timeout)
                if wait >= 0 or nativePk.valid:
                    break

            if not nativePk.valid:
                return None

            pk = CRTPPacket()
            pk.port = nativePk.port
            pk.channel = nativePk.channel
            pk.data = nativePk.payload
            return pk
        except Exception as e:
            if self._link_error_callback is not None:
                import traceback
                self._link_error_callback(
                    'Error communicating! Perhaps your device has been unplugged?\n'
                    'Exception:{}\n\n{}'.format(e, traceback.format_exc()))

    def get_status(self):
        """
        Return a status string from the interface.
        """
        'okay'

    def get_name(self):
        """
        Return a human readable name of the interface.
        """
        'cflinkcpp'

    def scan_interface(self, address=None):
        """
        Scan interface for available Crazyflie quadcopters and return a list
        with them.
        """
        if address:
            uris = cflinkcpp.Connection.scan(address)
        else:
            uris = cflinkcpp.Connection.scan()
        # convert to list of tuples, where the second part is a comment
        result = [(uri, '') for uri in uris]
        return result

    def scan_selected(self, uris):
        """
        Scan interface for available Crazyflie quadcopters and return a list
        with them.
        """
        return cflinkcpp.Connection.scan_selected(uris)

    def enum(self):
        """Enumerate, and return a list, of the available link URI on this
        system
        """
        return self.scan_interface()

    def get_help(self):
        """return the help message on how to form the URI for this driver
        None means no help
        """
        ''

    def close(self):
        """Close the link"""
        self._connection.close()
        self._connection = None

    def _recompute_link_quality_timer(self):
        if self._connection is None:
            return
        stats = self._connection.statistics
        sent_count = stats.sent_count - self._last_connection_stats.sent_count
        ack_count = stats.ack_count - self._last_connection_stats.ack_count
        if sent_count > 0:
            self._radio_link_statistics.link_quality = min(ack_count, sent_count) / sent_count * 100.0
        else:
            self._radio_link_statistics.link_quality = 1
        self._last_connection_stats = stats

        if self._radio_link_statistics_callback is not None:
            self._radio_link_statistics_callback(self._radio_link_statistics)

        if sent_count > 10 and ack_count == 0 and self._link_error_callback is not None:
            self._link_error_callback('Too many packets lost')

        threading.Timer(1.0, self._recompute_link_quality_timer).start()



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2021 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
import time
import unittest

from single_cf_grounded import TestSingleCfGrounded

from cflib.bootloader import Bootloader
from cflib.bootloader.boottypes import BootVersion


class TestBootloader(TestSingleCfGrounded):

    def test_boot_to_bootloader(self):
        self.assertTrue(self.is_stm_connected())
        bl = Bootloader(self.radioUri)
        started = bl.start_bootloader(warm_boot=True)
        self.assertTrue(started)

        # t = bl.get_target(TargetTypes.NRF51)
        # print(t)

        bl.reset_to_firmware()
        bl.close()
        time.sleep(1)
        self.assertTrue(self.is_stm_connected())
        self.assertTrue(BootVersion.is_cf2(bl.protocol_version))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
An early serial link driver. This could still be used (after some fixing) to
run high-speed CRTP with the Crazyflie. The UART can be run at 2Mbit.
"""
import logging
import queue
import re
import struct
import threading

from .crtpstack import CRTPPacket
from .exceptions import WrongUriType
from cflib.cpx import CPX
from cflib.cpx import CPXFunction
from cflib.cpx import CPXPacket
from cflib.cpx import CPXTarget
from cflib.cpx.transports import UARTTransport
from cflib.crtp.crtpdriver import CRTPDriver

found_serial = True
try:
    import serial.tools.list_ports as list_ports
except ImportError:
    found_serial = False

__author__ = 'Bitcraze AB'
__all__ = ['SerialDriver']

logger = logging.getLogger(__name__)


class SerialDriver(CRTPDriver):

    def __init__(self):
        CRTPDriver.__init__(self)
        self.ser = None
        self.uri = ''
        self.link_error_callback = None
        self.in_queue = None
        self.out_queue = None
        self._receive_thread = None
        self._send_thread = None
        self.needs_resending = False
        logger.info('Initialized serial driver.')

    def connect(self, uri, linkQualityCallback, linkErrorCallback):
        # check if the URI is a serial URI
        if not re.search('^serial://', uri):
            raise WrongUriType('Not a serial URI')

        # Check if it is a valid serial URI
        uri_data = re.search('^serial://([-a-zA-Z0-9/.]+)$', uri)
        if not uri_data:
            raise Exception('Invalid serial URI')

        # Move to Serial transport?
        device_name = uri_data.group(1)
        devices = self.get_devices()
        if device_name not in devices:
            raise Exception('Could not identify device')
        device = devices[device_name]

        self.uri = uri

        self.link_error_callback = linkErrorCallback

        # Prepare the inter-thread communication queue
        self.in_queue = queue.Queue()

        self.cpx = CPX(UARTTransport(device, 576000))

        self._thread = _CPXReceiveThread(self.cpx, self.in_queue,
                                         linkErrorCallback)
        self._thread.start()

        # Switch the link bridge to CPX in the Crazyflie
        self.cpx.sendPacket(CPXPacket(destination=CPXTarget.STM32,
                                      function=CPXFunction.SYSTEM,
                                      data=[0x21, 0x01]))
        # Force client connect to true
        self.cpx.sendPacket(CPXPacket(destination=CPXTarget.STM32,
                                      function=CPXFunction.SYSTEM,
                                      data=[0x20, 0x01]))

    def send_packet(self, pk):
        raw = (pk.header,) + struct.unpack('B' * len(pk.data), pk.data)
        self.cpx.sendPacket(CPXPacket(destination=CPXTarget.STM32,
                                      function=CPXFunction.CRTP,
                                      data=raw))

    def receive_packet(self, wait=0):
        try:
            if wait == 0:
                pk = self.in_queue.get(False)
            elif wait < 0:
                pk = self.in_queue.get(True)
            else:
                pk = self.in_queue.get(True, wait)
        except queue.Empty:
            return None
        return pk

    def get_status(self):
        return 'No information available'

    def get_name(self):
        return 'serial'

    def scan_interface(self, address):
        print('Scanning serial')
        if found_serial:
            print('Found serial')
            devices_names = self.get_devices().keys()
            print(devices_names)
            return [('serial://' + x, '') for x in devices_names]
        else:
            return []

    def close(self):
        """ Close the link. """
        # Stop the comm thread
        self._thread.stop()

        # Close the socket
        try:
            self.cpx.close()
            self.cpx = None

        except Exception as e:
            print(e)
            logger.error('Could not close {}'.format(e))
            pass
        print('Driver closed')

    def get_devices(self):
        result = {}
        for port in list_ports.comports():
            name = port.name
            # Name is not populated on all systems, fall back on the device
            if not name:
                name = port.device

            result[name] = port.device

        return result


class _CPXReceiveThread(threading.Thread):
    """
    Radio link receiver thread used to read data from the
    Socket. """

    def __init__(self, cpx, inQueue, link_error_callback):
        """ Create the object """
        threading.Thread.__init__(self)
        self._cpx = cpx
        self.in_queue = inQueue
        self.sp = False
        self.link_error_callback = link_error_callback

    def stop(self):
        """ Stop the thread """
        self.sp = True
        try:
            self.join()
        except Exception:
            pass

    def run(self):
        """ Run the receiver thread """

        while (True):
            if (self.sp):
                break
            try:
                # Block until a packet is available though the socket
                # CPX receive will only return full packets
                cpxPacket = self._cpx.receivePacket(CPXFunction.CRTP, timeout=1)
                data = struct.unpack('B' * cpxPacket.length, cpxPacket.data)
                if len(data) > 0:
                    pk = CRTPPacket(data[0],
                                    list(data[1:]))
                    self.in_queue.put(pk)
            except queue.Empty:
                pass  # This is ok
            except Exception as e:
                import traceback

                self.link_error_callback(
                    'Error communicating with the Crazyflie\n'
                    'Exception:%s\n\n%s' % (e,
                                            traceback.format_exc()))



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2019 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Simple example that connects to the crazyflie at `URI` and writes to
parameters that control the LED-ring,
it has been tested with (and designed for) the LED-ring deck.

Change the URI variable to your Crazyflie configuration.
"""
import logging
import time

import cflib.crtp
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.utils import uri_helper

# URI to the Crazyflie to connect to
URI = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')

# Only output errors from the logging framework
logging.basicConfig(level=logging.ERROR)


if __name__ == '__main__':
    # Initialize the low-level drivers
    cflib.crtp.init_drivers()

    with SyncCrazyflie(URI) as scf:
        cf = scf.cf

        # Set solid color effect
        cf.param.set_value('ring.effect', '7')
        # Set the RGB values
        cf.param.set_value('ring.solidRed', '100')
        cf.param.set_value('ring.solidGreen', '0')
        cf.param.set_value('ring.solidBlue', '0')
        time.sleep(2)

        # Set black color effect
        cf.param.set_value('ring.effect', '0')
        time.sleep(1)

        # Set fade to color effect
        cf.param.set_value('ring.effect', '14')
        # Set fade time i seconds
        cf.param.set_value('ring.fadeTime', '1.0')
        # Set the RGB values in one uint32 0xRRGGBB
        cf.param.set_value('ring.fadeColor', int('0000A0', 16))
        time.sleep(1)
        cf.param.set_value('ring.fadeColor', int('00A000', 16))
        time.sleep(1)
        cf.param.set_value('ring.fadeColor', int('A00000', 16))
        time.sleep(1)



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2025 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
The reset_estimator() function resets the Kalman filter and makes the
Crazyflie wait until it has an accurate position estimate. It is mostly
used with the Lighthouse Positioning System to ensure that the Crazyflie
won't take off before it has accurate positioning data.
"""
import time

from cflib.crazyflie.log import LogConfig
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.crazyflie.syncLogger import SyncLogger


def reset_estimator(crazyflie):
    if isinstance(crazyflie, SyncCrazyflie):
        cf = crazyflie.cf
    else:
        cf = crazyflie
    cf.param.set_value('kalman.resetEstimation', '1')
    time.sleep(0.1)
    cf.param.set_value('kalman.resetEstimation', '0')

    _wait_for_position_estimator(cf)


def _wait_for_position_estimator(cf):
    print('Waiting for estimator to find position...', end='\r')

    log_config = LogConfig(name='Kalman Variance', period_in_ms=500)
    log_config.add_variable('kalman.varPX', 'float')
    log_config.add_variable('kalman.varPY', 'float')
    log_config.add_variable('kalman.varPZ', 'float')

    var_y_history = [1000] * 10
    var_x_history = [1000] * 10
    var_z_history = [1000] * 10

    threshold = 0.001

    with SyncLogger(cf, log_config) as logger:
        for log_entry in logger:
            data = log_entry[1]

            var_x_history.append(data['kalman.varPX'])
            var_x_history.pop(0)
            var_y_history.append(data['kalman.varPY'])
            var_y_history.pop(0)
            var_z_history.append(data['kalman.varPZ'])
            var_z_history.pop(0)

            min_x = min(var_x_history)
            max_x = max(var_x_history)
            min_y = min(var_y_history)
            max_y = max(var_y_history)
            min_z = min(var_z_history)
            max_z = max(var_z_history)

            if (max_x - min_x) < threshold and (
                    max_y - min_y) < threshold and (
                    max_z - min_z) < threshold:
                print('Waiting for estimator to find position...success!')
                break



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2022 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import numpy as np

from cflib.localization.lighthouse_bs_vector import LighthouseBsVector
from cflib.localization.lighthouse_bs_vector import LighthouseBsVectors
from cflib.localization.lighthouse_types import LhDeck4SensorPositions
from cflib.localization.lighthouse_types import Pose


class LighthouseFixtures:
    """
    Fixtures to be used in lighthouse unit tests
    """

    # Stock objects
    # BS0 is pointing along the X-axis
    BS0_POSE = Pose(t_vec=(-2.0, 1.0, 3.0))
    # BS1 is pointing along the Y-axis
    BS1_POSE = Pose.from_rot_vec(R_vec=(0.0, 0.0, np.pi / 2), t_vec=(0.0, -2.0, 3.0))
    # BS2 is pointing along the negative Y-axis
    BS2_POSE = Pose.from_rot_vec(R_vec=(0.0, 0.0, -np.pi / 2), t_vec=(0.0, 2.0, 3.0))
    # BS3 is pointing along the negative X-axis
    BS3_POSE = Pose.from_rot_vec(R_vec=(0.0, 0.0, np.pi), t_vec=(2.0, 0.0, 2.0))

    # CF_ORIGIN is in the origin, pointing along the X-axis
    CF_ORIGIN_POSE = Pose()

    # CF1 is pointing along the X-axis
    CF1_POSE = Pose(t_vec=(0.3, 0.2, 0.1))

    # CF2 is pointing along the Y-axis
    CF2_POSE = Pose.from_rot_vec(R_vec=(0.0, 0.0, np.pi / 2), t_vec=(1.0, 0.0, 0.0))

    def __init__(self) -> None:
        self.angles_cf_origin_bs0 = self.synthesize_angles(self.CF_ORIGIN_POSE, self.BS0_POSE)
        self.angles_cf_origin_bs1 = self.synthesize_angles(self.CF_ORIGIN_POSE, self.BS1_POSE)

        self.angles_cf1_bs1 = self.synthesize_angles(self.CF1_POSE, self.BS1_POSE)
        self.angles_cf1_bs2 = self.synthesize_angles(self.CF1_POSE, self.BS2_POSE)

        self.angles_cf2_bs0 = self.synthesize_angles(self.CF2_POSE, self.BS0_POSE)
        self.angles_cf2_bs1 = self.synthesize_angles(self.CF2_POSE, self.BS1_POSE)
        self.angles_cf2_bs2 = self.synthesize_angles(self.CF2_POSE, self.BS2_POSE)
        self.angles_cf2_bs3 = self.synthesize_angles(self.CF2_POSE, self.BS3_POSE)

    def synthesize_angles(self, pose_cf: Pose, pose_bs: Pose) -> LighthouseBsVectors:
        """
        Genereate a LighthouseBsVectors object based
        """

        result = LighthouseBsVectors()
        for sens_pos_ref_cf in LhDeck4SensorPositions.positions:
            sens_pos_ref_global = pose_cf.rotate_translate(sens_pos_ref_cf)
            sens_pos_ref_bs = pose_bs.inv_rotate_translate(sens_pos_ref_global)
            result.append(LighthouseBsVector.from_cart(sens_pos_ref_bs))
        return result



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2019 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Example of how to read the Lighthouse base station geometry and
calibration memory from a Crazyflie
"""
import logging
from threading import Event

import cflib.crtp  # noqa
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.mem import LighthouseMemHelper
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.utils import uri_helper

# Only output errors from the logging framework
logging.basicConfig(level=logging.ERROR)


class ReadMem:
    def __init__(self, uri):
        self._event = Event()

        with SyncCrazyflie(uri, cf=Crazyflie(rw_cache='./cache')) as scf:
            helper = LighthouseMemHelper(scf.cf)

            helper.read_all_geos(self._geo_read_ready)
            self._event.wait()

            self._event.clear()

            helper.read_all_calibs(self._calib_read_ready)
            self._event.wait()

    def _geo_read_ready(self, geo_data):
        for id, data in geo_data.items():
            print('---- Geometry for base station', id + 1)
            data.dump()
            print()
        self._event.set()

    def _calib_read_ready(self, calib_data):
        for id, data in calib_data.items():
            print('---- Calibration data for base station', id + 1)
            data.dump()
            print()
        self._event.set()


if __name__ == '__main__':
    # URI to the Crazyflie to connect to
    uri = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')

    # Initialize the low-level drivers
    cflib.crtp.init_drivers()

    ReadMem(uri)



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2025 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import time

import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.utils import uri_helper

# URI to the Crazyflie to connect to
URI = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')


class rgb:
    def __init__(self, r: int, g: int, b: int):
        self.r = r
        self.g = g
        self.b = b


class hsv:
    def __init__(self, h: int, s: int, v: int):
        self.h = h
        self.s = s
        self.v = v


def hsv_to_rgb(input: hsv) -> rgb:
    h, s, v = input.h, input.s / 255.0, input.v / 255.0

    if s == 0:
        r = g = b = v
    else:
        h = h / 60.0
        i = int(h)
        f = h - i
        p = v * (1.0 - s)
        q = v * (1.0 - s * f)
        t = v * (1.0 - s * (1.0 - f))

        if i == 0:
            r, g, b = v, t, p
        elif i == 1:
            r, g, b = q, v, p
        elif i == 2:
            r, g, b = p, v, t
        elif i == 3:
            r, g, b = p, q, v
        elif i == 4:
            r, g, b = t, p, v
        else:
            r, g, b = v, p, q

    return rgb(int(r * 255), int(g * 255), int(b * 255))


def cycle_colors(step: int) -> rgb:
    h = step % 360
    s = 255
    v = 255
    return hsv_to_rgb(hsv(h, s, v))


def construct_uint32_color(input: rgb) -> int:
    return (input.r << 16) | (input.g << 8) | input.b


if __name__ == '__main__':
    # Initialize the low-level drivers
    cflib.crtp.init_drivers()

    with SyncCrazyflie(URI, cf=Crazyflie(rw_cache='./cache')) as scf:
        cf = scf.cf

        try:
            print('Cycling through colors. Press Ctrl-C to stop.')
            while True:
                for i in range(0, 360, 1):
                    color = cycle_colors(i)
                    color_uint32 = construct_uint32_color(color)
                    cf.param.set_value('led_deck_ctrl.rgb888', str(color_uint32))
                    time.sleep(0.01)
        except KeyboardInterrupt:
            print('\nStopping and turning off LED...')
            cf.param.set_value('led_deck_ctrl.rgb888', '0')
            time.sleep(0.1)



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2019 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
This example demonstrate the LEDTIMING led ring sequence memory. This memory and
led-ring effect allows to pre-program a LED sequence to be played autonomously
by the Crazyflie.

Change the URI variable to your Crazyflie configuration.
"""
import logging
import time

import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.mem import MemoryElement
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.utils import uri_helper

# URI to the Crazyflie to connect to
URI = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')

# Only output errors from the logging framework
logging.basicConfig(level=logging.ERROR)


if __name__ == '__main__':
    # Initialize the low-level drivers
    cflib.crtp.init_drivers()

    with SyncCrazyflie(URI, cf=Crazyflie(rw_cache='./cache')) as scf:
        cf = scf.cf

        # Get LED memory and write to it
        mem = cf.mem.get_mems(MemoryElement.TYPE_DRIVER_LEDTIMING)
        if len(mem) > 0:
            mem[0].add(25, {'r': 100, 'g': 0, 'b': 0})
            mem[0].add(0, {'r': 0, 'g': 100, 'b': 0}, leds=1)
            mem[0].add(0, {'r': 0, 'g': 100, 'b': 0}, leds=2)
            mem[0].add(3000, {'r': 0, 'g': 100, 'b': 0}, leds=3, rotate=1)
            mem[0].add(50, {'r': 0, 'g': 0, 'b': 100}, leds=1)
            mem[0].add(25, {'r': 0, 'g': 0, 'b': 100}, leds=0, fade=True)
            mem[0].add(25, {'r': 100, 'g': 0, 'b': 100}, leds=1)
            mem[0].add(25, {'r': 100, 'g': 0, 'b': 0})
            mem[0].add(50, {'r': 100, 'g': 0, 'b': 100})
            mem[0].write_data(None)
        else:
            print('No LED ring present')

        # Set virtual mem effect effect
        cf.param.set_value('ring.effect', '0')
        time.sleep(2)
        cf.param.set_value('ring.effect', '17')
        time.sleep(2)



# Eric Yihan Chen
# The Automatic Coordination of Teams (ACT) Lab
# University of Southern California
# ericyihc@usc.edu
'''
    Simple example that connects to the first Crazyflie found, triggers
    reading of rssi data and acknowledgement rate for every channel (0 to 125).
    It finally sets the Crazyflie channel back to default, plots link
    quality data, and offers good channel suggestion.

    This script should be used on a Crazyflie with bluetooth disabled and RSSI
    ack packet enabled to get RSSI feedback. To configure the Crazyflie in this
    mode build the crazyflie2-nrf-firmware with `CFLAGS += -DRSSI_ACK_PACKET=1`
    in `config.mk`. Additionally, the Crazyflie must be using the default address
    0xE7E7E7E7E7. See https://github.com/bitcraze/crazyflie-lib-python/issues/131
    for more information.
'''
import argparse

import matplotlib.pyplot as plt
import numpy as np

import cflib.drivers.crazyradio as crazyradio

radio = crazyradio.Crazyradio()

# optional user input
parser = argparse.ArgumentParser(description='Key variables')
parser.add_argument(
    '-try', '--try', dest='TRY', type=int, default=100,
    help='the time to send data for each channel'
)
# by default my crazyflie uses channel 80
parser.add_argument(
    '-channel', '--channel', dest='channel', type=int,
    default=80, help='the default channel in crazyflie'
)
# by default my crazyflie uses datarate 2M
parser.add_argument(
    '-rate', '--rate', dest='rate', type=int, default=2,
    help='the default datarate in crazyflie'
)
parser.add_argument(
    '-frac', '--fraction',  dest='fraction', type=float,
    default=0.25, help='top fraction of suggested channels'
)
args = parser.parse_args()

init_channel = args.channel
TRY = args.TRY
Fraction = args.fraction
data_rate = args.rate

radio.set_channel(init_channel)
radio.set_data_rate(data_rate)
SET_RADIO_CHANNEL = 1

rssi_std = []
rssi = []
ack = []
radio.set_arc(0)

for channel in range(0, 126, 1):

    # change Crazyflie channel
    for x in range(50):
        radio.send_packet((0xff, 0x03, SET_RADIO_CHANNEL, channel))

    count = 0
    temp = []

    # change radio channel
    radio.set_channel(channel)

    for i in range(TRY):
        pk = radio.send_packet((0xff, ))
        if pk.ack:
            count += 1
        if pk.ack and len(pk.data) > 2 and \
           pk.data[0] & 0xf3 == 0xf3 and pk.data[1] == 0x01:
            # append rssi data
            temp.append(pk.data[2])

    ack_rate = count / TRY
    if len(temp) > 0:
        rssi_avg = np.mean(temp)
        std = np.std(temp)
    else:
        rssi_avg = np.nan
        std = np.nan

    rssi.append(rssi_avg)
    ack.append(ack_rate)
    rssi_std.append(std)

    print('Channel', channel, 'ack_rate:', ack_rate,
          'rssi average:', rssi_avg, 'rssi std:', std)

# change channel back to default
for x in range(50):
    radio.send_packet((0xff, 0x03, SET_RADIO_CHANNEL, init_channel))

# divide each std by 2 for plotting convenience
rssi_std = [x / 2 for x in rssi_std]
rssi_std = np.array(rssi_std)
rssi = np.array(rssi)
ack = np.array(ack)

rssi_rank = []
ack_rank = []

# suggestion for rssi
order = rssi.argsort()
ranks = order.argsort()
for x in range(int(125 * Fraction)):
    for y in range(126):
        if ranks[y] == x:
            rssi_rank.append(y)

# suggestion for ack
order = ack.argsort()
ranks = order.argsort()
for x in range(126, 126 - int(125 * Fraction) - 1, -1):
    for y in range(126):
        if ranks[y] == x:
            ack_rank.append(y)

rssi_set = set(rssi_rank[0:int(125 * Fraction)])
ack_set = set(ack_rank[0:int(125 * Fraction)])
final_rank = rssi_set.intersection(ack_rank)
print('\nSuggested Channels:')
for x in final_rank:
    print('\t', x)

# graph 1 for ack
x = np.arange(0, 126, 1)
fig, ax1 = plt.subplots()
ax1.axis([0, 125, 0, 1.25])
ax1.plot(x, ack, 'b')
ax1.set_xlabel('Channel')
ax1.set_ylabel('Ack Rate', color='b')
for tl in ax1.get_yticklabels():
    tl.set_color('b')

# graph 2 for rssi & rssi_std
ax2 = ax1.twinx()
ax2.grid(True)
ax2.errorbar(x, rssi, yerr=rssi_std, fmt='r-')
ax2.fill_between(x, rssi + rssi_std, rssi - rssi_std,
                 facecolor='orange', edgecolor='k')
ax2.axis([0, 125, 0, 90])
plt.ylabel('RSSI Average', color='r')
for tl in ax2.get_yticklabels():
    tl.set_color('r')
points = np.ones(100)
for x in final_rank:
    ax2.plot((x, x), (0, 100), linestyle='-',
             color='cornflowerblue', linewidth=1)

plt.show()



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2025 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Simple script for visualizing the Ligithouse positioning system's configuration
using matplotlib. Each base station is represented by a local coordinate frame, while
each one's coverage is represented by 2 circular sectors; a horizontal and a vertical one.
Notice that the base station coordinate frame is defined as:
    - X-axis pointing forward through the glass
    - Y-axis pointing right, when the base station is seen from the front.
    - Z-axis pointing up

To run the script, just change the path to your .yaml file.
"""
import matplotlib.pyplot as plt
import numpy as np
import yaml

config_file = 'lighthouse.yaml'  # Add the path to your .yaml file

Range = 5  # Range of each base station in meters
FoV_h = 150  # Horizontal Field of View in degrees
FoV_v = 110  # Vertical Field of View in degrees


def draw_coordinate_frame(ax, P, R, label='', length=0.5, is_bs=False):
    """Draw a coordinate frame at position t with orientation R."""
    x_axis = R @ np.array([length, 0, 0])
    y_axis = R @ np.array([0, length, 0])
    z_axis = R @ np.array([0, 0, length])

    ax.quiver(P[0], P[1], P[2], x_axis[0], x_axis[1], x_axis[2], color='r', linewidth=2)
    ax.quiver(P[0], P[1], P[2], y_axis[0], y_axis[1], y_axis[2], color='g', linewidth=2)
    ax.quiver(P[0], P[1], P[2], z_axis[0], z_axis[1], z_axis[2], color='b', linewidth=2)
    if is_bs:
        ax.scatter(P[0], P[1], P[2], s=50, color='black')
    ax.text(P[0], P[1], P[2], label, fontsize=10, color='black')


def draw_horizontal_sector(ax, P, R, radius=Range, angle_deg=FoV_h, color='r', alpha=0.3, n_points=50):
    """
    Draw a circular sector centered at the origin of the local coordinate frame,lying in
    the local XY-plane, so that its central axis is aligned with the positive X-axis.
    """
    # Angle range (centered on X-axis)
    half_angle = np.deg2rad(angle_deg / 2)
    thetas = np.linspace(-half_angle, half_angle, n_points)

    # Circle points in local XY-plane
    x_local = radius * np.cos(thetas)
    y_local = radius * np.sin(thetas)
    z_local = np.zeros_like(thetas)

    # Stack the coordinates into a 3xN matix
    pts_local = np.vstack([x_local, y_local, z_local])

    # Transfer the points to the global frame, creating a 3xN matrix
    pts_global = R @ pts_local + P.reshape(3, 1)

    # Close the sector by adding the center point at the start and end
    X = np.concatenate(([P[0]], pts_global[0, :], [P[0]]))
    Y = np.concatenate(([P[1]], pts_global[1, :], [P[1]]))
    Z = np.concatenate(([P[2]], pts_global[2, :], [P[2]]))

    # Plot filled sector
    ax.plot_trisurf(X, Y, Z, color=color, alpha=alpha, linewidth=0)


def draw_vertical_sector(ax, P, R, radius=Range, angle_deg=FoV_v, color='r', alpha=0.3, n_points=50):
    """
    Draw a circular sector centered at the origin of the local coordinate frame,lying in
    the local XZ-plane, so that its central axis is aligned with the positive X-axis.
    """
    # Angle range (centered on X-axis)
    half_angle = np.deg2rad(angle_deg / 2)
    thetas = np.linspace(-half_angle, half_angle, n_points)

    # Circle points in local XZ-plane
    x_local = radius * np.cos(thetas)
    y_local = np.zeros_like(thetas)
    z_local = radius * np.sin(thetas)

    # Stack the coordinates into a 3xN matix
    pts_local = np.vstack([x_local, y_local, z_local])

    # Transfer the points to the global frame, creating a 3xN matrix
    pts_global = R @ pts_local + P.reshape(3, 1)

    # Close the sector by adding the center point at the start and end
    X = np.concatenate(([P[0]], pts_global[0, :], [P[0]]))
    Y = np.concatenate(([P[1]], pts_global[1, :], [P[1]]))
    Z = np.concatenate(([P[2]], pts_global[2, :], [P[2]]))

    # Plot filled sector
    ax.plot_trisurf(X, Y, Z, color=color, alpha=alpha, linewidth=0)


if __name__ == '__main__':
    # Load the .yamnl file
    with open(config_file, 'r') as f:
        data = yaml.safe_load(f)
    geos = data['geos']

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    # Draw global frame
    draw_coordinate_frame(ax, np.zeros(3), np.eye(3), label='Global', length=1.0)

    # Draw local frames + sectors
    for key, geo in geos.items():
        origin = np.array(geo['origin'])
        rotation = np.array(geo['rotation'])
        draw_coordinate_frame(ax, origin, rotation, label=f'BS {key+1}', length=0.5, is_bs=True)

        # Local XY-plane sector
        draw_horizontal_sector(ax, origin, rotation, radius=Range, angle_deg=FoV_h, color='red', alpha=0.15)

        # Local YZ-plane sector
        draw_vertical_sector(ax, origin, rotation, radius=Range, angle_deg=FoV_v, color='red', alpha=0.15)

    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')
    ax.set_title('Lighthouse Visualization')

    # Set equal aspect ratio
    all_points = [np.array(geo['origin']) for geo in geos.values()]
    all_points.append(np.zeros(3))
    all_points = np.array(all_points)
    max_range = np.ptp(all_points, axis=0).max()
    mid = all_points.mean(axis=0)
    ax.set_xlim(mid[0] - max_range/2, mid[0] + max_range/2)
    ax.set_ylim(mid[1] - max_range/2, mid[1] + max_range/2)
    ax.set_zlim(mid[2] - max_range/2, mid[2] + max_range/2)

    plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
USB driver for the Crazyflie.
"""
import logging
import os
import platform

import libusb_package
import usb
import usb.core

try:
    if os.environ['CRTP_PCAP_LOG'] is not None:
        from cflib.crtp.pcap import PCAPLog
except KeyError:
    pass

__author__ = 'Bitcraze AB'
__all__ = ['CfUsb']

logger = logging.getLogger(__name__)

USB_VID = 0x0483
USB_PID = 0x5740


def _find_devices():
    """
    Returns a list of CrazyRadio devices currently connected to the computer
    """
    ret = []

    logger.info('Looking for devices....')

    if os.name == 'nt':
        import usb.backend.libusb0 as libusb0

        backend = libusb0.get_backend()
    else:
        backend = libusb_package.get_libusb1_backend()

    devices = usb.core.find(idVendor=USB_VID, idProduct=USB_PID, find_all=1,
                            backend=backend)
    if devices:
        for d in devices:
            if d.manufacturer == 'Bitcraze AB':
                ret.append(d)

    return ret


class CfUsb:
    """ Used for communication with the Crazyradio USB dongle """

    def __init__(self, device=None, devid=0):
        """ Create object and scan for USB dongle if no device is supplied """
        self.dev = None
        self.handle = None
        self._last_write = 0
        self._last_read = 0

        if device is None:
            devices = _find_devices()
            try:
                self.dev = devices[devid]
            except Exception:
                self.dev = None

        try:  # configuration might already be confgiured by composite VCP, try claim interface
            usb.util.claim_interface(self.dev, 0)
        except Exception:
            try:
                self.dev.set_configuration()  # it was not, then set configuration
            except Exception:
                if self.dev:
                    if platform.system() == 'Linux':
                        self.dev.reset()
                        self.dev.set_configuration()

        self.handle = self.dev
        if self.dev:
            self.version = float('{0:x}.{1:x}'.format(self.dev.bcdDevice >> 8, self.dev.bcdDevice & 0x0FF))
        else:
            self.version = 0.0

    def get_serial(self):
        # The signature for get_string has changed between versions to 1.0.0b1,
        # 1.0.0b2 and 1.0.0. Try the old signature first, if that fails try
        # the newer one.
        try:
            return usb.util.get_string(self.dev, 255, self.dev.iSerialNumber)
        except (usb.core.USBError, ValueError):
            return usb.util.get_string(self.dev, self.dev.iSerialNumber)

    def close(self):
        if self.dev:
            usb.util.dispose_resources(self.dev)

        self.handle = None
        self.dev = None

    def scan(self):
        # TODO: Currently only supports one device
        if self.dev:
            return [('usb://0', '')]
        return []

    def set_crtp_to_usb(self, crtp_to_usb: bool):
        if crtp_to_usb:
            _send_vendor_setup(self.handle, 0x01, 0x01, 1, ())
        else:
            _send_vendor_setup(self.handle, 0x01, 0x01, 0, ())

    def _log_packet(self, receive, id, packet):
        try:
            if os.environ['CRTP_PCAP_LOG'] is not None:
                if len(packet) > 0:
                    logger = PCAPLog.instance()
                    logger.logCRTP(
                        logger.LinkType.USB,
                        receive,
                        id,
                        bytearray.fromhex(self.get_serial()),
                        0,
                        packet
                    )
        except KeyError:
            pass

    # Data transfers
    def send_packet(self, dataOut):
        """ Send a packet and receive the ack from the radio dongle
            The ack contains information about the packet transmission
            and a data payload if the ack packet contained any """
        try:
            self.handle.write(endpoint=1, data=dataOut, timeout=20)
            self._log_packet(False, self.dev.port_number, dataOut)
        except usb.USBError:
            pass

    def receive_packet(self):
        dataIn = ()
        try:
            dataIn = self.handle.read(0x81, 64, timeout=20)
        except usb.USBError as e:
            try:
                if e.backend_error_code == -7 or e.backend_error_code == -116:
                    # Normal, the read was empty
                    pass
                else:
                    raise IOError('Crazyflie disconnected')
            except AttributeError:
                # pyusb < 1.0 doesn't implement getting the underlying error
                # number and it seems as if it's not possible to detect
                # if the cable is disconnected. So this detection is not
                # supported, but the "normal" case will work.
                pass

        self._log_packet(True, self.dev.port_number, dataIn)

        return dataIn


# Private utility functions
def _send_vendor_setup(handle, request, value, index, data):
    handle.ctrl_transfer(usb.TYPE_VENDOR, request, wValue=value,
                         wIndex=index, timeout=1000, data_or_wLength=data)


def _get_vendor_setup(handle, request, value, index, length):
    return handle.ctrl_transfer(usb.TYPE_VENDOR | 0x80, request,
                                wValue=value, wIndex=index, timeout=1000,
                                data_or_wLength=length)



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2016-2020 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
import unittest
from test.support.asyncCallbackCaller import AsyncCallbackCaller
from unittest.mock import call
from unittest.mock import MagicMock

from cflib.crazyflie import Crazyflie
from cflib.crazyflie.log import Log
from cflib.crazyflie.log import LogConfig
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.crazyflie.syncLogger import SyncLogger
from cflib.utils.callbacks import Caller


class SyncLoggerTest(unittest.TestCase):

    def setUp(self):
        self.cf_mock = MagicMock(spec=Crazyflie)
        self.cf_mock.disconnected = Caller()

        self.log_mock = MagicMock(spec=Log)
        self.cf_mock.log = self.log_mock

        self.log_config_mock = MagicMock(spec=LogConfig)
        self.log_config_mock.data_received_cb = Caller()

        self.log_config_mock2 = MagicMock(spec=LogConfig)
        self.log_config_mock2.data_received_cb = Caller()

        self.sut = SyncLogger(self.cf_mock, self.log_config_mock)
        self.sut_multi = SyncLogger(
            self.cf_mock, [self.log_config_mock, self.log_config_mock2])

    def test_that_log_configuration_is_added_on_connect(self):
        # Fixture

        # Test
        self.sut.connect()

        # Assert
        self.log_mock.add_config.assert_called_once_with(self.log_config_mock)

    def test_that_multiple_log_configurations_are_added_on_connect(self):
        # Fixture

        # Test
        self.sut_multi.connect()

        # Assert
        self.log_mock.add_config.assert_has_calls([
            call(self.log_config_mock),
            call(self.log_config_mock2)
        ])

    def test_that_logging_is_started_on_connect(self):
        # Fixture

        # Test
        self.sut.connect()

        # Assert
        self.log_config_mock.start.assert_called_once_with()

    def test_that_logging_is_started_on_connect_for_multiple_log_confs(self):
        # Fixture

        # Test
        self.sut_multi.connect()

        # Assert
        self.log_config_mock.start.assert_called_once_with()
        self.log_config_mock2.start.assert_called_once_with()

    def test_connection_status_after_connect(self):
        # Fixture
        self.sut.connect()

        # Test
        actual = self.sut.is_connected()

        # Assert
        self.assertTrue(actual)

    def test_that_callbacks_are_removed_on_disconnect(self):
        # Fixture

        # Test
        self.sut.connect()
        self.sut.disconnect()

        # Assert
        self.assertEqual(0, len(self.cf_mock.disconnected.callbacks))
        self.assertEqual(0,
                         len(self.log_config_mock.data_received_cb.callbacks))

    def test_that_log_config_is_stopped_on_disconnect(self):
        # Fixture
        self.sut.connect()

        # Test
        self.sut.disconnect()

        # Assert
        self.log_config_mock.stop.assert_called_once_with()
        self.log_config_mock.delete.assert_called_once_with()

    def test_that_multiple_log_configs_are_stopped_on_disconnect(self):
        # Fixture
        self.sut_multi.connect()

        # Test
        self.sut_multi.disconnect()

        # Assert
        self.log_config_mock.stop.assert_called_once_with()
        self.log_config_mock.delete.assert_called_once_with()

        self.log_config_mock2.stop.assert_called_once_with()
        self.log_config_mock2.delete.assert_called_once_with()

    def test_that_data_is_received(self):
        # Fixture
        self.sut.connect()

        expected = ('Some ts', 'Some data', 'Some logblock')
        AsyncCallbackCaller(cb=self.log_config_mock.data_received_cb,
                            args=[expected[0], expected[1], expected[2]]
                            ).trigger()

        # Test
        actual = None
        for log_block in self.sut:
            actual = log_block
            break

        # Assert
        self.assertEqual(expected, actual)

    def test_connection_status_after_disconnect(self):
        # Fixture
        self.sut.connect()
        self.sut.disconnect()

        # Test
        actual = self.sut.is_connected()

        # Assert
        self.assertFalse(actual)

    def test_that_connect_to_connected_instance_raises_exception(self):
        # Fixture
        self.sut.connect()

        # Test
        # Assert
        with self.assertRaises(Exception):
            self.sut.connect()

    def test_connect_to_disconnected_instance(self):
        # Fixture
        self.sut.connect()
        self.sut.disconnect()

        # Test
        self.sut.connect()

        # Assert
        # Nothing here. Just not expecting an exception

    def test_disconnect_from_disconnected_instance(self):
        # Fixture

        # Test
        self.sut.disconnect()

        # Assert
        # Nothing here. Just not expecting an exception

    def test_connect_disconnect_with_context_management(self):
        # Fixture

        # Test
        with (SyncLogger(self.cf_mock, self.log_config_mock)) as sut:
            # Assert
            self.assertTrue(sut.is_connected())

        self.assertFalse(sut.is_connected())

    def test_that_iterator_is_implemented(self):
        # Fixture

        # Test
        actual = self.sut.__iter__()

        # Assert
        self.assertEqual(self.sut, actual)

    def test_construction_with_sync_crazyflie_instance(self):
        # Fixture
        scf_mock = MagicMock(spec=SyncCrazyflie)
        scf_mock.cf = self.cf_mock

        # Test
        sut = SyncLogger(scf_mock, self.log_config_mock)
        sut.connect()

        # Assert
        # Nothing here. Just not expecting an exception

    def test_getting_data_without_connection_raises_exception(self):
        # Fixture

        # Test
        with self.assertRaises(StopIteration):
            self.sut.__next__()

            # Assert

    def test_lost_connection_in_crazyflie_disconnects(self):
        # Fixture
        self.sut.connect()

        # Test
        AsyncCallbackCaller(cb=self.cf_mock.disconnected,
                            args=['Some uri']
                            ).call_and_wait()

        # Assert
        self.assertFalse(self.sut.is_connected())

    def test_lost_connection_in_crazyflie_raises_exception_in_iterator(self):
        # Fixture
        self.sut.connect()

        # Note this is not foolproof, the disconnected callback may be called
        # before we are waiting on data. It will raise the same exception
        # though and will pass
        AsyncCallbackCaller(cb=self.cf_mock.disconnected,
                            delay=0.5,
                            args=['Some uri']
                            ).trigger()

        # Test
        # Assert
        with self.assertRaises(StopIteration):
            self.sut.__next__()



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2024 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import time

import cflib.crtp  # noqa
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.utils import uri_helper

# Reads the CFLIB_URI environment variable for URI or uses default
uri = uri_helper.uri_from_env(default='radio://0/90/2M/E7E7E7E7E7')


def latency_callback(latency: float):
    """A callback to run when we get an updated latency estimate"""
    print(f'Latency: {latency:.3f} ms')


if __name__ == '__main__':
    # Initialize the low-level drivers
    cflib.crtp.init_drivers()

    # Create Crazyflie object, with cache to avoid re-reading param and log TOC
    cf = Crazyflie(rw_cache='./cache')

    # Add a callback to whenever we receive an updated latency estimate
    #
    # This could also be a Python lambda, something like:
    cf.link_statistics.latency.latency_updated.add_callback(latency_callback)

    # This will connect the Crazyflie with the URI specified above.
    with SyncCrazyflie(uri, cf=cf) as scf:
        print('[host] Connected, use ctrl-c to quit.')

        while True:
            time.sleep(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
CRTP Driver main class.
"""

__author__ = 'Bitcraze AB'
__all__ = ['CRTPDriver']


class CRTPDriver:
    """ CTRP Driver main class

    This class in inherited by all the CRTP link drivers.
    """

    def __init__(self):
        """Driver constructor. Throw an exception if the driver is unable to
        open the URI
        """
        self.needs_resending = True

    def connect(self, uri, radio_link_statistics_callback, link_error_callback):
        """Connect the driver to a specified URI

        @param uri Uri of the link to open
        @param radio_link_statistics_callback Callback to report radio link statistics
        @param link_error_callback Callback to report errors (will result in
               disconnection)
        """

    def send_packet(self, pk):
        """Send a CRTP packet"""

    def receive_packet(self, wait=0):
        """Receive a CRTP packet.

        @param wait The time to wait for a packet in second. -1 means forever

        @return One CRTP packet or None if no packet has been received.
        """

    def get_status(self):
        """
        Return a status string from the interface.
        """

    def get_name(self):
        """
        Return a human readable name of the interface.
        """

    def scan_interface(self, address=None):
        """
        Scan interface for available Crazyflie quadcopters and return a list
        with them.
        """

    def enum(self):
        """Enumerate, and return a list, of the available link URI on this
        system
        """

    def get_help(self):
        """return the help message on how to form the URI for this driver
        None means no help
        """

    def close(self):
        """Close the link"""



import collections
import time

import numpy as np


class RadioLinkStatistics:
    """
    Tracks the health of the signal by monitoring link quality, uplink RSSI,
    packet rates, and congestion.
    """

    def __init__(self, radio_link_statistics_callback, alpha=0.1):
        """
        Initialize the RadioLinkStatistics class.

        :param alpha: Weight for the exponential moving average (default 0.1)
        """
        self._alpha = alpha
        self._radio_link_statistics_callback = radio_link_statistics_callback

        self._retries = collections.deque()
        self._retry_sum = 0

    def update(self, ack, packet_out):
        """
        Update the radio link statistics based on the acknowledgment data.

        :param ack: Acknowledgment object containing retry and RSSI data.
        """
        self.radio_link_statistics = {}

        self._update_link_quality(ack)
        self._update_rssi(ack)
        self._update_rate_and_congestion(ack, packet_out)

        if self.radio_link_statistics and self._radio_link_statistics_callback:
            self._radio_link_statistics_callback(self.radio_link_statistics)

    def _update_link_quality(self, ack):
        """
        Updates the link quality based on the number of retries.

        :param ack: Acknowledgment object with retry data.
        """
        if ack:
            retry = 10 - ack.retry
            self._retries.append(retry)
            self._retry_sum += retry
            if len(self._retries) > 100:
                self._retry_sum -= self._retries.popleft()
            self.radio_link_statistics['link_quality'] = float(self._retry_sum) / len(self._retries) * 10

    def _update_rssi(self, ack):
        """
        Updates the uplink RSSI based on the acknowledgment signal.

        :param ack: Acknowledgment object with RSSI data.
        """
        if not hasattr(self, '_rssi_timestamps'):
            self._rssi_timestamps = collections.deque(maxlen=100)
        if not hasattr(self, '_rssi_values'):
            self._rssi_values = collections.deque(maxlen=100)

        # update RSSI if the acknowledgment contains RSSI data
        if ack.ack and len(ack.data) > 2 and ack.data[0] & 0xf3 == 0xf3 and ack.data[1] == 0x01:
            instantaneous_rssi = ack.data[2]
            self._rssi_values.append(instantaneous_rssi)
            self._rssi_timestamps.append(time.time())

            # Calculate time-weighted average RSSI
            if len(self._rssi_timestamps) >= 2:  # At least 2 points are needed to calculate differences
                time_diffs = np.diff(self._rssi_timestamps, prepend=time.time())
                weights = np.exp(-time_diffs)
                weighted_average = np.sum(weights * self._rssi_values) / np.sum(weights)
                self.radio_link_statistics['uplink_rssi'] = weighted_average

    def _update_rate_and_congestion(self, ack, packet_out):
        """
        Updates the packet rate and bandwidth congestion based on the acknowledgment data.

        :param ack: Acknowledgment object with congestion data.
        """
        if not hasattr(self, '_previous_time_stamp'):
            self._previous_time_stamp = time.time()
        if not hasattr(self, '_amount_null_packets_up'):
            self._amount_null_packets_up = 0
        if not hasattr(self, '_amount_packets_up'):
            self._amount_packets_up = 0
        if not hasattr(self, '_amount_null_packets_down'):
            self._amount_null_packets_down = 0
        if not hasattr(self, '_amount_packets_down'):
            self._amount_packets_down = 0

        self._amount_packets_up += 1  # everytime this function is called, a packet is sent
        if not packet_out:  # if the packet is empty, we send a null packet
            self._amount_null_packets_up += 1

        # Find null packets in the downlink and count them
        mask = 0b11110011
        if ack.data:
            empty_ack_packet = int(ack.data[0]) & mask

            if empty_ack_packet == 0xF3:
                self._amount_null_packets_down += 1
            self._amount_packets_down += 1

            # rate and congestion stats every N seconds
            if time.time() - self._previous_time_stamp > 0.1:
                # self._uplink_rate = self._amount_packets_up / (time.time() - self._previous_time_stamp)
                self.radio_link_statistics['uplink_rate'] = self._amount_packets_up / \
                    (time.time() - self._previous_time_stamp)
                self.radio_link_statistics['downlink_rate'] = self._amount_packets_down / \
                    (time.time() - self._previous_time_stamp)
                self.radio_link_statistics['uplink_congestion'] = 1.0 - \
                    self._amount_null_packets_up / self._amount_packets_up
                self.radio_link_statistics['downlink_congestion'] = 1.0 - \
                    self._amount_null_packets_down / self._amount_packets_down

                self._amount_packets_up = 0
                self._amount_null_packets_up = 0
                self._amount_packets_down = 0
                self._amount_null_packets_down = 0
                self._previous_time_stamp = time.time()



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2022 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import copy

import numpy as np
import numpy.typing as npt

from cflib.localization.lighthouse_bs_vector import LighthouseBsVector
from cflib.localization.lighthouse_types import LhCfPoseSample
from cflib.localization.lighthouse_types import Pose


class LighthouseSystemScaler:
    """This class is used to re-scale a system based on various measurements."""
    @classmethod
    def scale_fixed_point(cls, bs_poses: dict[int, Pose], cf_poses: list[Pose], expected: npt.ArrayLike,
                          actual: Pose) -> tuple[dict[int, Pose], list[Pose], float]:
        """
        Scale a system based on a position in the physical world in relation to where it is in the estimated system
        geometry. Assume the system is aligned and simply use the distance to the points for scaling.

        :param bs_poses: a dictionary with the base station poses in the current reference frame
        :param cf_poses: List of CF poses
        :param expected: The real world position to use as reference
        :param actual: The estimated position in the current system geometry
        :return: a tuple containing a dictionary with the base station poses in the scaled system,
                 a list of Crazyflie poses in the scaled system and the scaling factor
        """
        expected_distance = np.linalg.norm(expected)
        actual_distance = np.linalg.norm(actual.translation)
        scale_factor = expected_distance / actual_distance
        return cls._scale_system(bs_poses, cf_poses, scale_factor)

    @classmethod
    def scale_diagonals(cls, bs_poses: dict[int, Pose], cf_poses: list[Pose], matched_samples: list[LhCfPoseSample],
                        expected_diagonal: float) -> tuple[dict[int, Pose], list[Pose], float]:
        """
        Scale a system based on where base station "rays" intersects the lighthouse deck in relation to sensor
        positions. Calculates the intersection points for all samples and scales the system to match the expected
        distance between sensors on the deck.

        :param bs_poses: a dictionary with the base station poses in the current reference frame
        :param cf_poses: List of CF poses
        :param matched_samples: List of samples. Length must be the same as cf_poses.
        :return: a tuple containing a dictionary with the base station poses in the scaled system,
                 a list of Crazyflie poses in the scaled system and the scaling factor
        """

        estimated_diagonal = cls._calculate_mean_diagonal(bs_poses, cf_poses, matched_samples)
        scale_factor = expected_diagonal / estimated_diagonal
        return cls._scale_system(bs_poses, cf_poses, scale_factor)

    @classmethod
    def _scale_system(cls, bs_poses: dict[int, Pose], cf_poses: list[Pose],
                      scale_factor: float) -> tuple[dict[int, Pose], list[Pose], float]:
        """
        Scale poses of base stations and crazyflie samples.
        """
        bs_scaled = {bs_id: copy.copy(pose) for bs_id, pose in bs_poses.items()}
        for pose in bs_scaled.values():
            pose.scale(scale_factor)

        cf_scaled = [copy.copy(pose) for pose in cf_poses]
        for pose in cf_scaled:
            pose.scale(scale_factor)

        return bs_scaled, cf_scaled, scale_factor

    @classmethod
    def _calculate_mean_diagonal(cls, bs_poses: dict[int, Pose], cf_poses: list[Pose],
                                 matched_samples: list[LhCfPoseSample]) -> float:
        """
        Calculate the average diagonal sensor distance based on where the rays intersect the lighthouse deck
        """
        diagonals: list[float] = []

        for cf_pose, sample in zip(cf_poses, matched_samples):
            for bs_id, vectors in sample.angles_calibrated.items():
                diagonals.append(cls.calc_intersection_distance(vectors[0], vectors[3], bs_poses[bs_id], cf_pose))
                diagonals.append(cls.calc_intersection_distance(vectors[1], vectors[2], bs_poses[bs_id], cf_pose))

        estimated_diagonal = np.mean(diagonals)

        return estimated_diagonal

    @classmethod
    def calc_intersection_distance(cls, vector1: LighthouseBsVector, vector2: LighthouseBsVector,
                                   bs_pose: Pose, cf_pose: Pose) -> float:
        """Calculate distance between intersection points of rays on the plane defined by the lighthouse deck"""

        intersection1 = cls.calc_intersection_point(vector1, bs_pose, cf_pose)
        intersection2 = cls.calc_intersection_point(vector2, bs_pose, cf_pose)
        distance = np.linalg.norm(intersection1 - intersection2)
        return distance

    @classmethod
    def calc_intersection_point(cls, vector: LighthouseBsVector, bs_pose: Pose, cf_pose: Pose) -> npt.NDArray:
        """Calculate the intersetion point of a lines and a plane.
        The line is the intersection of the two light planes from a base station, while the
        plane is defined by the lighthouse deck of the Crazyflie."""

        plane_base = cf_pose.translation
        plane_normal = np.dot(cf_pose.rot_matrix, (0.0, 0.0, 1.0))

        line_base = bs_pose.translation
        line_vector = np.dot(bs_pose.rot_matrix, vector.cart)

        dist_on_line = np.dot((plane_base - line_base), plane_normal) / np.dot(line_vector, plane_normal)

        return line_base + line_vector * dist_on_line



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2019 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Simple example that connects to one crazyflie, sets the initial position/yaw
and flies a trajectory.

The initial pose (x, y, z, yaw) is configured in a number of variables and
the trajectory is flown relative to this position, using the initial yaw.

This example is intended to work with any absolute positioning system.
It aims at documenting how to take off with the Crazyflie in an orientation
that is different from the standard positive X orientation and how to set the
initial position of the kalman estimator.
"""
import math
import time

import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.utils import uri_helper
from cflib.utils.reset_estimator import reset_estimator

# URI to the Crazyflie to connect to
uri = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')

# Change the sequence according to your setup
#             x    y    z
sequence = [
    (0, 0, 0.7),
    (-0.7, 0, 0.7),
    (0, 0, 0.7),
    (0, 0, 0.2),
]


def set_initial_position(scf, x, y, z, yaw_deg):
    scf.cf.param.set_value('kalman.initialX', x)
    scf.cf.param.set_value('kalman.initialY', y)
    scf.cf.param.set_value('kalman.initialZ', z)

    yaw_radians = math.radians(yaw_deg)
    scf.cf.param.set_value('kalman.initialYaw', yaw_radians)


def run_sequence(scf, sequence, base_x, base_y, base_z, yaw):
    cf = scf.cf

    # Arm the Crazyflie
    cf.platform.send_arming_request(True)
    time.sleep(1.0)

    for position in sequence:
        print('Setting position {}'.format(position))

        x = position[0] + base_x
        y = position[1] + base_y
        z = position[2] + base_z

        for i in range(50):
            cf.commander.send_position_setpoint(x, y, z, yaw)
            time.sleep(0.1)

    cf.commander.send_stop_setpoint()
    # Hand control over to the high level commander to avoid timeout and locking of the Crazyflie
    cf.commander.send_notify_setpoint_stop()

    # Make sure that the last packet leaves before the link is closed
    # since the message queue is not flushed before closing
    time.sleep(0.1)


if __name__ == '__main__':
    cflib.crtp.init_drivers()

    # Set these to the position and yaw based on how your Crazyflie is placed
    # on the floor
    initial_x = 1.0
    initial_y = 1.0
    initial_z = 0.0
    initial_yaw = 90  # In degrees
    # 0: positive X direction
    # 90: positive Y direction
    # 180: negative X direction
    # 270: negative Y direction

    with SyncCrazyflie(uri, cf=Crazyflie(rw_cache='./cache')) as scf:
        set_initial_position(scf, initial_x, initial_y, initial_z, initial_yaw)
        reset_estimator(scf)
        run_sequence(scf, sequence,
                     initial_x, initial_y, initial_z, initial_yaw)



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2016 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Simple example that connects to one crazyflie (check the address at the top
and update it to your crazyflie address) and send a sequence of setpoints,
one every 5 seconds.

This example is intended to work with the Loco Positioning System in TWR TOA
mode and with the Lighthouse Positioning System. It aims at documenting how
to set the Crazyflie in position control mode and how to send setpoints.
"""
import time

import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.log import LogConfig
from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.utils import uri_helper
from cflib.utils.reset_estimator import reset_estimator

# URI to the Crazyflie to connect to
uri = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')

# Change the sequence according to your setup
#             x    y    z  YAW
sequence = [
    (0.0, 0.0, 0.4, 0),
    (0.0, 0.0, 1.2, 0),
    (0.5, -0.5, 1.2, 0),
    (0.5, 0.5, 1.2, 0),
    (-0.5, 0.5, 1.2, 0),
    (-0.5, -0.5, 1.2, 0),
    (0.0, 0.0, 1.2, 0),
    (0.0, 0.0, 0.4, 0),
    (0.0, 0.0, 0.0, 0),
]


def take_off(cf, position):
    take_off_time = 1.0
    sleep_time = 0.1
    steps = int(take_off_time / sleep_time)
    vz = position[2] / take_off_time

    print(f'take off at {position[2]}')

    for i in range(steps):
        cf.commander.send_velocity_world_setpoint(0, 0, vz, 0)
        time.sleep(sleep_time)


def position_callback(timestamp, data, logconf):
    x = data['kalman.stateX']
    y = data['kalman.stateY']
    z = data['kalman.stateZ']
    print('pos: ({}, {}, {})'.format(x, y, z))


def start_position_printing(scf):
    log_conf = LogConfig(name='Position', period_in_ms=500)
    log_conf.add_variable('kalman.stateX', 'float')
    log_conf.add_variable('kalman.stateY', 'float')
    log_conf.add_variable('kalman.stateZ', 'float')

    scf.cf.log.add_config(log_conf)
    log_conf.data_received_cb.add_callback(position_callback)
    log_conf.start()


def run_sequence(scf, sequence):
    cf = scf.cf

    # Arm the Crazyflie
    cf.platform.send_arming_request(True)
    time.sleep(1.0)

    take_off(cf, sequence[0])
    time.sleep(1.0)

    for position in sequence:
        print('Setting position {}'.format(position))
        for i in range(50):
            cf.commander.send_position_setpoint(position[0],
                                                position[1],
                                                position[2],
                                                position[3])
            time.sleep(0.1)

    cf.commander.send_stop_setpoint()
    # Hand control over to the high level commander to avoid timeout and locking of the Crazyflie
    cf.commander.send_notify_setpoint_stop()

    # Make sure that the last packet leaves before the link is closed
    # since the message queue is not flushed before closing
    time.sleep(0.1)


if __name__ == '__main__':
    cflib.crtp.init_drivers()

    with SyncCrazyflie(uri, cf=Crazyflie(rw_cache='./cache')) as scf:
        reset_estimator(scf)
        start_position_printing(scf)
        run_sequence(scf, sequence)



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2014 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Simple example that connects to the first Crazyflie found, logs the Stabilizer
and prints it to the console. After 10s the application disconnects and exits.
"""
import logging
import time
from threading import Timer

import cflib.crtp  # noqa
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.log import LogConfig
from cflib.utils import uri_helper

uri = uri_helper.uri_from_env(default='radio://0/80/2M/E7E7E7E7E7')

# Only output errors from the logging framework
logging.basicConfig(level=logging.ERROR)


class LoggingExample:
    """
    Simple logging example class that logs the Stabilizer from a supplied
    link uri and disconnects after 5s.
    """

    def __init__(self, link_uri):
        """ Initialize and run the example with the specified link_uri """

        self._cf = Crazyflie(rw_cache='./cache')

        # Connect some callbacks from the Crazyflie API
        self._cf.connected.add_callback(self._connected)
        self._cf.disconnected.add_callback(self._disconnected)
        self._cf.connection_failed.add_callback(self._connection_failed)
        self._cf.connection_lost.add_callback(self._connection_lost)

        print('Connecting to %s' % link_uri)

        # Try to connect to the Crazyflie
        self._cf.open_link(link_uri)

        # Variable used to keep main loop occupied until disconnect
        self.is_connected = True

    def _connected(self, link_uri):
        """ This callback is called form the Crazyflie API when a Crazyflie
        has been connected and the TOCs have been downloaded."""
        print('Connected to %s' % link_uri)

        # The definition of the logconfig can be made before connecting
        self._lg_stab = LogConfig(name='Stabilizer', period_in_ms=100)
        self._lg_stab.add_variable('stateEstimate.x', 'float')
        self._lg_stab.add_variable('stateEstimate.y', 'float')
        self._lg_stab.add_variable('stateEstimate.z', 'float')
        self._lg_stab.add_variable('stabilizer.roll', 'float')
        self._lg_stab.add_variable('stabilizer.pitch', 'float')
        self._lg_stab.add_variable('stabilizer.yaw', 'float')
        # The fetch-as argument can be set to FP16 to save space in the log packet
        self._lg_stab.add_variable('pm.vbat', 'FP16')

        # Adding the configuration cannot be done until a Crazyflie is
        # connected, since we need to check that the variables we
        # would like to log are in the TOC.
        try:
            self._cf.log.add_config(self._lg_stab)
            # This callback will receive the data
            self._lg_stab.data_received_cb.add_callback(self._stab_log_data)
            # This callback will be called on errors
            self._lg_stab.error_cb.add_callback(self._stab_log_error)
            # Start the logging
            self._lg_stab.start()
        except KeyError as e:
            print('Could not start log configuration,'
                  '{} not found in TOC'.format(str(e)))
        except AttributeError:
            print('Could not add Stabilizer log config, bad configuration.')

        # Start a timer to disconnect in 10s
        t = Timer(5, self._cf.close_link)
        t.start()

    def _stab_log_error(self, logconf, msg):
        """Callback from the log API when an error occurs"""
        print('Error when logging %s: %s' % (logconf.name, msg))

    def _stab_log_data(self, timestamp, data, logconf):
        """Callback from a the log API when data arrives"""
        print(f'[{timestamp}][{logconf.name}]: ', end='')
        for name, value in data.items():
            print(f'{name}: {value:3.3f} ', end='')
        print()

    def _connection_failed(self, link_uri, msg):
        """Callback when connection initial connection fails (i.e no Crazyflie
        at the specified address)"""
        print('Connection to %s failed: %s' % (link_uri, msg))
        self.is_connected = False

    def _connection_lost(self, link_uri, msg):
        """Callback when disconnected after a connection has been made (i.e
        Crazyflie moves out of range)"""
        print('Connection to %s lost: %s' % (link_uri, msg))

    def _disconnected(self, link_uri):
        """Callback when the Crazyflie is disconnected (called in all cases)"""
        print('Disconnected from %s' % link_uri)
        self.is_connected = False


if __name__ == '__main__':
    # Initialize the low-level drivers
    cflib.crtp.init_drivers()

    le = LoggingExample(uri)

    # The Crazyflie lib doesn't contain anything to keep the application alive,
    # so this is where your application should do something. In our case we
    # are just waiting until we are disconnected.
    while le.is_connected:
        time.sleep(1)



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2024 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
from threading import Event

from cflib.crazyflie import Crazyflie
from cflib.localization.param_io import ParamFileManager


class ParamFileHelper:
    '''ParamFileHelper is a helper to synchonously write multiple paramteters
    from a file and store them in persistent memory'''

    def __init__(self, crazyflie):
        if isinstance(crazyflie, Crazyflie):
            self._cf = crazyflie
            self.persistent_sema = None
            self.success = False
        else:
            raise TypeError('ParamFileHelper only takes a Crazyflie Object')

    def _persistent_stored_callback(self, complete_name, success):
        self.success = success
        if not success:
            print(f'Persistent params: failed to store {complete_name}!')
        else:
            print(f'Persistent params: stored {complete_name}!')
        self.persistent_sema.set()

    def store_params_from_file(self, filename):
        params = ParamFileManager().read(filename)
        for param, state in params.items():
            self.persistent_sema = Event()
            self._cf.param.set_value(param, state.stored_value)
            self._cf.param.persistent_store(param, self._persistent_stored_callback)
            self.persistent_sema.wait()
            if not self.success:
                break
        return self.success



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2018-2022 Bitcraze AB
#
#  Crazyflie Python Library
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
"""
Example script which can be used to fly the Crazyflie in "FPV" mode
using the Flow deck and the AI deck.

The example shows how to connect to a Crazyflie over the WiFi and
use both CPX and CRTP for communication over WiFI.

When the application is started the Crazyflie will hover at 0.3 m. The
Crazyflie can then be controlled by using keyboard input:
 * Move by using the arrow keys (left/right/forward/backwards)
 * Adjust the right with w/s (0.1 m for each keypress)
 * Yaw slowly using a/d (CCW/CW)
 * Yaw fast using z/x (CCW/CW)

The demo is ended by closing the application.

For the example to run the following hardware is needed:
 * Crazyflie 2.1
 * Crazyradio PA
 * Flow v2 deck
 * AI deck 1.1
"""
import logging
import struct
import sys
import threading
from time import time

import numpy as np

import cflib.crtp
from cflib.cpx import CPXFunction
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.log import LogConfig
from cflib.utils import uri_helper

try:
    from sip import setapi
    setapi('QVariant', 2)
    setapi('QString', 2)
except ImportError:
    pass

from PyQt6 import QtCore, QtWidgets, QtGui

logging.basicConfig(level=logging.INFO)

# If you have set a CFLIB_URI environment variable, that address will be used.
URI = uri_helper.uri_from_env(default='tcp://192.168.4.1:5000')
# 192.168.4.1 is the default IP address if the aideck is Access point.
# If you are using the aideck as a station, you should use the assigned IP address
# 5000 is the default port for the aideck

CAM_HEIGHT = 244
CAM_WIDTH = 324
# Set the speed factor for moving and rotating
SPEED_FACTOR = 0.3

if len(sys.argv) > 1:
    URI = sys.argv[1]


class ImageDownloader(threading.Thread):
    def __init__(self, cpx, cb):
        threading.Thread.__init__(self)
        self.daemon = True
        self._cpx = cpx
        self._cb = cb

    def run(self):
        while True:
            p = self._cpx.receivePacket(CPXFunction.APP)
            [magic, width, height, depth, format, size] = struct.unpack('<BHHBBI', p.data[0:11])
            if (magic == 0xBC):
                imgStream = bytearray()
                while len(imgStream) < size:
                    p = self._cpx.receivePacket(CPXFunction.APP)
                    imgStream.extend(p.data)

                bayer_img = np.frombuffer(imgStream, dtype=np.uint8)
                self._cb(bayer_img)


class MainWindow(QtWidgets.QWidget):

    def __init__(self, URI):
        QtWidgets.QWidget.__init__(self)

        self.setWindowTitle('Crazyflie / AI deck FPV demo')

        self.mainLayout = QtWidgets.QVBoxLayout()

        self.image_frame = QtWidgets.QLabel()
        self.mainLayout.addWidget(self.image_frame)

        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.addWidget(QtWidgets.QLabel('Position (X/Y/Z)'), 0, 0, 1, 1, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.gridLayout.addWidget(QtWidgets.QLabel('Pose (roll/pitch/yaw)'), 1,
                                  0, 1, 1, QtCore.Qt.AlignmentFlag.AlignLeft)

        self.labels = {
            'stateEstimate.x': {
                'widget': QtWidgets.QLabel('X'),
                'x_grid': 0, 'y_grid': 1,
                'alignment': QtCore.Qt.AlignmentFlag.AlignLeft
            },
            'stateEstimate.y': {
                'widget': QtWidgets.QLabel('Y'),
                'x_grid': 0, 'y_grid': 2,
                'alignment': QtCore.Qt.AlignmentFlag.AlignLeft
            },
            'stateEstimate.z': {
                'widget': QtWidgets.QLabel('Z'),
                'x_grid': 0, 'y_grid': 3,
                'alignment': QtCore.Qt.AlignmentFlag.AlignLeft
            },
            'stabilizer.roll': {
                'widget': QtWidgets.QLabel('roll'),
                'x_grid': 1, 'y_grid': 1,
                'alignment': QtCore.Qt.AlignmentFlag.AlignLeft
            },
            'stabilizer.pitch': {
                'widget': QtWidgets.QLabel('pitch'),
                'x_grid': 1, 'y_grid': 2,
                'alignment': QtCore.Qt.AlignmentFlag.AlignLeft
            },
            'stabilizer.yaw': {
                'widget': QtWidgets.QLabel('yaw'),
                'x_grid': 1, 'y_grid': 3,
                'alignment': QtCore.Qt.AlignmentFlag.AlignLeft
            }
        }

        for name in self.labels:
            w = self.labels[name]
            self.gridLayout.addWidget(w['widget'], w['x_grid'], w['y_grid'], 1, 1, w['alignment'])

        self.mainLayout.addLayout(self.gridLayout)

        self.setLayout(self.mainLayout)

        cflib.crtp.init_drivers()
        self.cf = Crazyflie(ro_cache=None, rw_cache='cache')

        # Connect callbacks from the Crazyflie API
        self.cf.connected.add_callback(self.connected)
        self.cf.disconnected.add_callback(self.disconnected)

        # Connect to the Crazyflie
        self.cf.open_link(URI)

        if not self.cf.link:
            print('Could not connect to Crazyflie')
            sys.exit(1)

        if not hasattr(self.cf.link, 'cpx'):
            print('Not connecting with WiFi')
            self.cf.close_link()
        else:
            self._imgDownload = ImageDownloader(self.cf.link.cpx, self.updateImage)
            self._imgDownload.start()

            # Arm the Crazyflie
            self.cf.platform.send_arming_request(True)
            time.sleep(1.0)

            self.hover = {'x': 0.0, 'y': 0.0, 'z': 0.0, 'yaw': 0.0, 'height': 0.3}

            self.hoverTimer = QtCore.QTimer()
            self.hoverTimer.timeout.connect(self.sendHoverCommand)
            self.hoverTimer.setInterval(100)
            self.hoverTimer.start()

    def updateImage(self, image):
        i = QtGui.QImage(image, CAM_WIDTH, CAM_HEIGHT, QtGui.QImage.Format_Grayscale8).scaled(324*2, 244*2)
        self.image_frame.setPixmap(QtGui.QPixmap.fromImage(i))

    def keyPressEvent(self, event):
        if (not event.isAutoRepeat()):
            if (event.key() == QtCore.Qt.Key.Key_Left):
                self.updateHover('y', 1)
            if (event.key() == QtCore.Qt.Key.Key_Right):
                self.updateHover('y', -1)
            if (event.key() == QtCore.Qt.Key.Key_Up):
                self.updateHover('x', 1)
            if (event.key() == QtCore.Qt.Key.Key_Down):
                self.updateHover('x', -1)
            if (event.key() == QtCore.Qt.Key.Key_A):
                self.updateHover('yaw', -70)
            if (event.key() == QtCore.Qt.Key.Key_D):
                self.updateHover('yaw', 70)
            if (event.key() == QtCore.Qt.Key.Key_Z):
                self.updateHover('yaw', -200)
            if (event.key() == QtCore.Qt.Key.Key_X):
                self.updateHover('yaw', 200)
            if (event.key() == QtCore.Qt.Key.Key_W):
                self.updateHover('height', 0.1)
            if (event.key() == QtCore.Qt.Key.Key_S):
                self.updateHover('height', -0.1)

    def keyReleaseEvent(self, event):
        if (not event.isAutoRepeat()):
            if (event.key() == QtCore.Qt.Key.Key_Left):
                self.updateHover('y', 0)
            if (event.key() == QtCore.Qt.Key.Key_Right):
                self.updateHover('y', 0)
            if (event.key() == QtCore.Qt.Key.Key_Up):
                self.updateHover('x', 0)
            if (event.key() == QtCore.Qt.Key.Key_Down):
                self.updateHover('x', 0)
            if (event.key() == QtCore.Qt.Key.Key_A):
                self.updateHover('yaw', 0)
            if (event.key() == QtCore.Qt.Key.Key_D):
                self.updateHover('yaw', 0)
            if (event.key() == QtCore.Qt.Key.Key_W):
                self.updateHover('height', 0)
            if (event.key() == QtCore.Qt.Key.Key_S):
                self.updateHover('height', 0)
            if (event.key() == QtCore.Qt.Key.Key_Z):
                self.updateHover('yaw', 0)
            if (event.key() == QtCore.Qt.Key.Key_X):
                self.updateHover('yaw', 0)

    def sendHoverCommand(self):
        self.cf.commander.send_hover_setpoint(
            self.hover['x'], self.hover['y'], self.hover['yaw'],
            self.hover['height'])

    def updateHover(self, k, v):
        if (k != 'height'):
            self.hover[k] = v * SPEED_FACTOR
        else:
            self.hover[k] += v

    def disconnected(self, URI):
        print('Disconnected')
        sys.exit(1)

    def connected(self, URI):
        print('We are now connected to {}'.format(URI))

        # The definition of the logconfig can be made before connecting
        lp = LogConfig(name='Position', period_in_ms=100)
        lp.add_variable('stateEstimate.x')
        lp.add_variable('stateEstimate.y')
        lp.add_variable('stateEstimate.z')
        lp.add_variable('stabilizer.roll')
        lp.add_variable('stabilizer.pitch')
        lp.add_variable('stabilizer.yaw')

        try:
            self.cf.log.add_config(lp)
            lp.data_received_cb.add_callback(self.pos_data)
            lp.start()
        except KeyError as e:
            print('Could not start log configuration,'
                  '{} not found in TOC'.format(str(e)))
        except AttributeError:
            print('Could not add Position log config, bad configuration.')

    def pos_data(self, timestamp, data, logconf):
        for name in data:
            self.labels[name]['widget'].setText('{:.02f}'.format(data[name]))

    def closeEvent(self, event):
        if (self.cf is not None):
            self.cf.close_link()


if __name__ == '__main__':
    appQt = QtWidgets.QApplication(sys.argv)
    win = MainWindow(URI)
    win.show()
    appQt.exec()



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2017 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
""" The LPS lib is an API that is used to communicate with Loco Positioning
anchors.

The main purpose of the lib is to manage LPP (Loco Positioning Protocol).

Initially it will use a Crazyflie with a Loco Positioning deck as a bridge to
transfer information to LoPo anchors, but may in the future use other means
of transportation.

Example:
cf = Crazyflie()
cf.open_link("radio://0/125")

anchor = LoPoAnchor(cf)
anchor.set_position(1, (1.23, 4.56, 7.89))

"""
